import io

import numpy as np
import pytest

from plateau.serialization.testing import BINARY_COLUMNS, get_dataframe_not_nested


class FakeStore:
    def __init__(self):
        self._brain = {}

    def put(self, key, value):
        self._brain[key] = value

    def get(self, key):
        return self._brain[key]

    def open(self, key):
        return io.BytesIO(self._brain[key])


@pytest.fixture
def store(tmpdir):
    return FakeStore()


@pytest.fixture(params=BINARY_COLUMNS)
def binary_value(request):
    return request.param


@pytest.fixture
def dataframe_not_nested():
    return get_dataframe_not_nested(10)


@pytest.fixture(autouse=True)
def numpy_errstate():
    with np.errstate(all="raise"):
        yield



from plateau.core.factory import DatasetFactory


def test_repr(store_factory):
    factory = DatasetFactory(
        dataset_uuid="dataset_uuid",
        store_factory=store_factory,  # does not exist
    )
    assert repr(factory) == "<DatasetFactory: uuid=dataset_uuid is_loaded=False>"



import contextlib
import datetime
import re
from datetime import date
from unittest import mock
from warnings import catch_warnings, simplefilter

import hypothesis.extra.numpy as hyp_np
import hypothesis.strategies as hyp_st
import numpy as np
import pandas as pd
from hypothesis.errors import NonInteractiveExampleWarning

from plateau.core.uuid import gen_uuid_object

TIME_TO_FREEZE = datetime.datetime(2000, 1, 1, 1, 1, 1, 1)
TIME_TO_FREEZE_ISO = "2000-01-01T01:01:01.000001"
TIME_TO_FREEZE_ISO_QUOTED = "2000-01-01T01%3A01%3A01.000001"


def get_dataframe_alltypes():
    """Return a pandas DataFrame of length one with a column for each commonly
    used data types."""
    # fmt: off
    not_nested = get_dataframe_not_nested()
    nested_types = pd.DataFrame(
        {
            "array_int8": pd.Series([np.array([1], dtype=np.int8)], dtype=object),
            "array_int16": pd.Series([np.array([1], dtype=np.int16)], dtype=object),
            "array_int32": pd.Series([np.array([1], dtype=np.int32)], dtype=object),
            "array_int64": pd.Series([np.array([1], dtype=np.int64)], dtype=object),
            "array_uint8": pd.Series([np.array([1], dtype=np.uint8)], dtype=object),
            "array_uint16": pd.Series([np.array([1], dtype=np.uint16)], dtype=object),
            "array_uint32": pd.Series([np.array([1], dtype=np.uint32)], dtype=object),
            "array_uint64": pd.Series([np.array([1], dtype=np.uint64)], dtype=object),
            "array_float32": pd.Series([np.array([1], dtype=np.float32)], dtype=object),
            "array_float64": pd.Series([np.array([1], dtype=np.float64)], dtype=object),
            "array_unicode": pd.Series([np.array(["Ö"], dtype=object)], dtype=object),
        }
    )

    return pd.concat([not_nested, nested_types], axis=1).reset_index(drop=True).sort_index(axis=1)
    # fmt: on


def get_dataframe_not_nested():
    return pd.DataFrame(
        {
            "bool": pd.Series([1], dtype=np.bool_),
            "int8": pd.Series([1], dtype=np.int8),
            "int16": pd.Series([1], dtype=np.int16),
            "int32": pd.Series([1], dtype=np.int32),
            "int64": pd.Series([1], dtype=np.int64),
            "uint8": pd.Series([1], dtype=np.uint8),
            "uint16": pd.Series([1], dtype=np.uint16),
            "uint32": pd.Series([1], dtype=np.uint32),
            "uint64": pd.Series([1], dtype=np.uint64),
            "float32": pd.Series([1.0], dtype=np.float32),
            "float64": pd.Series([1.0], dtype=np.float64),
            "date": pd.Series([date(2018, 1, 1)], dtype=object),
            "datetime64": pd.Series(["2018-01-01"], dtype="datetime64[ns]"),
            "unicode": pd.Series(["Ö"], dtype=str),
            "null": pd.Series([None], dtype=object),
            # Adding a byte type with value as byte sequence which can not be encoded as UTF8
            "byte": pd.Series([gen_uuid_object().bytes], dtype=object),
        }
    ).sort_index(axis=1)


def get_scalar_dtype_strategy(exclude=None):
    """A `hypothesis` strategy yielding."""
    possible_strategies = {
        "datetime": hyp_np.datetime64_dtypes(max_period="ms", min_period="ns"),
        "uint": hyp_np.unsigned_integer_dtypes(),
        "int": hyp_np.integer_dtypes(),
        "float": hyp_np.floating_dtypes(),
        "byte": hyp_np.byte_string_dtypes(),
        "unicode": hyp_np.unicode_string_dtypes(),
    }
    if exclude is None:
        exclude = {}
    elif not isinstance(exclude, list):
        exclude = [exclude]
    for ex in exclude:
        if ex in possible_strategies:
            del possible_strategies[ex]
        else:
            raise ValueError(
                f"Strategy {ex} unknown. Possible values are {possible_strategies.keys()}"
            )
    return hyp_st.one_of(*list(possible_strategies.values()))


def get_numpy_array_strategy(
    shape=10, exclude_dtypes=None, unique=False, sort=False, allow_nan=True
):
    # the text example generation has quite some overhead when called the first time.
    # we don't want this in our test sample generation since the HealthCheck of hypothesis
    # might be triggered.
    with catch_warnings():
        simplefilter("ignore", NonInteractiveExampleWarning)
        hyp_st.text().example()

    dtype_strategy = get_scalar_dtype_strategy(exclude_dtypes)
    array_strategy = hyp_np.arrays(dtype=dtype_strategy, shape=shape, unique=unique)

    if exclude_dtypes is None or "date" not in exclude_dtypes:
        date_start = hyp_st.lists(
            hyp_st.dates(
                min_value=datetime.date(1970, 1, 1), max_value=datetime.date(2100, 1, 1)
            ),
            min_size=shape,
            max_size=shape,
            unique=unique,
        )
        one_of_strategies = [array_strategy] + [date_start.map(np.array)]
        array_strategy = hyp_st.one_of(one_of_strategies)

    def _restrict_datetime_ranges(arr):
        if np.issubdtype(arr.dtype, np.datetime64):
            return all(
                (arr < np.datetime64("2200-01-01"))
                & (arr > np.datetime64("1970-01-01"))
            )
        return True

    if exclude_dtypes is None or "datetime" not in exclude_dtypes:
        array_strategy = array_strategy.filter(_restrict_datetime_ranges)

    # Don't check for issues with surrogates, we don't handle them.
    def _restrict_surrogate_strings(arr):
        has_surrogate = re.compile("[\ud800-\udbff,\udc00-\udfff]", re.UNICODE)
        search_surrogate = np.vectorize(lambda x: bool(has_surrogate.search(x)))
        if np.issubdtype(arr.dtype, np.str_):
            return not any(search_surrogate(arr))
        return True

    if exclude_dtypes is None or "unicode" not in exclude_dtypes:
        array_strategy = array_strategy.filter(_restrict_surrogate_strings)

    if not allow_nan:

        def _check_for_nan(arr):
            if np.issubdtype(arr.dtype, np.floating):
                return not any(np.isnan(arr))
            return True

        array_strategy = array_strategy.filter(_check_for_nan)
    if unique and allow_nan:

        def _maximum_single_nan(arr):
            if np.issubdtype(arr.dtype, np.floating):
                return sum(np.isnan(arr)) <= 1
            return True

        array_strategy = array_strategy.filter(_maximum_single_nan)

    if sort:
        array_strategy = array_strategy.map(np.sort)
    return array_strategy


@contextlib.contextmanager
def cm_frozen_time(time_to_freeze):
    """Context manager to monkeypatch plateau.core._time.* to return a fixed
    datetime value `time_to_freeze`."""
    with (
        mock.patch("plateau.core._time.datetime_now") as mock_now,
        mock.patch("plateau.core._time.datetime_utcnow") as mock_utcnow,
    ):
        mock_now.return_value = time_to_freeze
        mock_utcnow.return_value = time_to_freeze
        yield



from .utils import create_dataset


def test_delete_dataset(store_factory, metadata_version, bound_delete_dataset):
    """Ensure that a dataset can be deleted."""
    create_dataset("dataset", store_factory, metadata_version)

    store = store_factory()
    assert len(list(store.keys())) > 0
    bound_delete_dataset("dataset", store_factory)
    assert len(list(store.keys())) == 0


def test_delete_single_dataset(store_factory, metadata_version, bound_delete_dataset):
    """Ensure that only the specified dataset is deleted."""
    create_dataset("dataset", store_factory, metadata_version)
    create_dataset("another_dataset", store_factory, metadata_version)
    store = store_factory()
    amount_of_keys = len(list(store.keys()))
    assert len(list(store.keys())) > 0
    bound_delete_dataset("dataset", store_factory)
    assert len(list(store.keys())) == amount_of_keys / 2, store.keys()


def test_delete_only_dataset(store_factory, metadata_version, bound_delete_dataset):
    """Ensure that files including the UUID but not starting with it are not
    deleted."""
    create_dataset("UUID", store_factory, metadata_version)

    store = store_factory()
    store.put(key="prefixUUID", data=b"")
    bound_delete_dataset("UUID", store_factory)
    assert "prefixUUID" in store.keys()


def test_delete_missing_dataset(store_factory, store_factory2, bound_delete_dataset):
    """Ensure that a dataset can be deleted even though some keys are already
    removed."""
    metadata_version = 4
    create_dataset("dataset", store_factory, metadata_version)

    store = store_factory()
    keys = sorted(store.keys())
    assert len(keys) > 0

    store2 = store_factory2()

    for missing in keys:
        if missing == "dataset.by-dataset-metadata.json":
            continue

        for k in keys:
            if k != missing:
                store2.put(k, store.get(k))

        bound_delete_dataset("dataset", store_factory2)
        assert len(list(store2.keys())) == 0


def test_delete_dataset_unreferenced_files(
    store_factory, metadata_version, bound_delete_dataset
):
    """Ensure that unreferenced files of a dataset are also removed when a
    dataset is deleted."""
    uuid = "dataset"
    create_dataset(uuid, store_factory, metadata_version)

    store = store_factory()
    store.put(f"{uuid}/table/trash.parquet", b"trash")

    assert len(list(store.keys())) > 0
    bound_delete_dataset(uuid, store_factory)

    assert len(list(store.keys())) == 0



"""Pandas performance helpers."""

from collections import OrderedDict
from typing import TYPE_CHECKING, Any, cast

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    import numpy.typing as npt

__all__ = (
    "aggregate_to_lists",
    "concat_dataframes",
    "drop_sorted_duplicates_keep_last",
    "is_dataframe_sorted",
    "mask_sorted_duplicates_keep_last",
    "merge_dataframes_robust",
    "sort_dataframe",
)


def concat_dataframes(dfs, default=None):
    """Concatenate given DataFrames.

    For non-empty iterables, this is roughly equivalent to::

        pd.concat(dfs, ignore_index=True, sort=False)

    except that the resulting index is undefined.

    .. important::

        If ``dfs`` is a list, it gets emptied during the process.


    .. warning::

        This requires all DataFrames to have the very same set of columns!

    Parameters
    ----------
    dfs: Iterable[pandas.DataFrame]
        Iterable of DataFrames w/ identical columns.
    default: Optional[pandas.DataFrame]
        Optional default if iterable is empty.

    Returns
    -------
    df: pandas.DataFrame
        Concatenated DataFrame or default value.

    Raises
    ------
    ValueError
        If iterable is empty but no default was provided.
    """
    # collect potential iterators
    if not isinstance(dfs, list):
        dfs = list(dfs)

    if len(dfs) == 0:
        if default is not None:
            res = default
        else:
            raise ValueError("Cannot concatenate 0 dataframes.")
    elif len(dfs) == 1:
        # that's faster than pd.concat w/ a single DF
        res = dfs[0]
    else:
        # pd.concat seems to hold the data in memory 3 times (not twice as you might expect it from naive copying the
        # input blocks into the output DF). This is very unfortunate especially for larger queries. This column-based
        # approach effectively reduces the maximum memory consumption and to our knowledge is not measuable slower.
        colset = set(dfs[0].columns)
        if not all(colset == set(df.columns) for df in dfs):
            raise ValueError("Not all DataFrames have the same set of columns!")

        res = pd.DataFrame(index=pd.RangeIndex(sum(len(df) for df in dfs)))
        for col in dfs[0].columns:
            res[col] = pd.concat(
                [df[col] for df in dfs], ignore_index=True, sort=False, copy=False
            )

    # ensure list (which is still referenced in parent scope) gets emptied
    del dfs[:]

    return res


def is_dataframe_sorted(df, columns):
    """Check that the given DataFrame is sorted as specified.

    This is more efficient than sorting the DataFrame.

    An empty DataFrame (no rows) is considered to be sorted.

    .. warning::

        This function does NOT handle NULL values correctly!

    Parameters
    ----------
    df: pd.DataFrame
        DataFrame to check.
    columns: Iterable[str]
        Column that the DataFrame should be sorted by.

    Returns
    -------
    sorted: bool
        ``True`` if DataFrame is sorted, ``False`` otherwise.

    Raises
    ------
    ValueError: If ``columns`` is empty.
    KeyError: If specified columns in ``by`` is missing.
    """
    columns = list(columns)

    if len(columns) == 0:
        raise ValueError("`columns` must contain at least 1 column")

    state = None

    for col in columns[::-1]:
        data = df[col].values
        if isinstance(data, pd.Categorical):
            data = np.asarray(data)
        data0 = data[:-1]
        data1 = data[1:]

        with np.errstate(invalid="ignore"):
            comp_le = data0 < data1
            comp_eq = data0 == data1

        if state is None:
            # last column
            state = comp_le | comp_eq
        else:
            state = comp_le | (comp_eq & state)

    return cast(Any, state).all()


def sort_dataframe(df, columns):
    """Sort DataFrame by columns.

    This is roughly equivalent to::

        df.sort_values(columns).reset_index(drop=True)

    .. warning::

        This function does NOT handle NULL values correctly!

    Parameters
    ----------
    df: pandas.DataFrame
        DataFrame to sort.
    columns: Iterable[str]
        Columns to sort by.

    Returns
    -------
    df: pandas.DataFrame
        Sorted DataFrame w/ reset index.
    """
    columns = list(columns)
    if is_dataframe_sorted(df, columns):
        return df
    data = [df[col].values for col in columns[::-1]]
    df = df.iloc[np.lexsort(data)]
    # reset inplace to reduce the memory usage
    df.reset_index(drop=True, inplace=True)
    return df


def mask_sorted_duplicates_keep_last(df, columns):
    """Mask duplicates on sorted data, keep last occurrence as unique entry.

    Roughly equivalent to::

        df.duplicated(subset=columns, keep='last').values

    .. warning:
        NULL-values are not supported!

    .. warning:
        The behavior on unsorted data is undefined!

    Parameters
    ----------
    df: pandas.DataFrame
        DataFrame in question.
    columns: Iterable[str]
        Column-subset for duplicate-check (remaining columns are ignored).

    Returns
    -------
    mask: numpy.ndarray
        1-dimensional boolean array, marking duplicates w/ ``True``
    """
    columns = list(columns)
    rows = len(df)
    mask: npt.NDArray[np.bool_] = np.zeros(rows, dtype=bool)

    if (rows > 1) and columns:
        sub: npt.NDArray[np.bool_] = np.ones(rows - 1, dtype=bool)
        for col in columns:
            data = df[col].values
            sub &= data[:-1] == data[1:]
        mask[:-1] = sub

    return mask


def drop_sorted_duplicates_keep_last(df, columns):
    """Drop duplicates on sorted data, keep last occurrence as unique entry.

    Roughly equivalent to::

        df.drop_duplicates(subset=columns, keep='last')

    .. warning:
        NULL-values are not supported!

    .. warning:
        The behavior on unsorted data is undefined!

    Parameters
    ----------
    df: pandas.DataFrame
        DataFrame in question.
    columns: Iterable[str]
        Column-subset for duplicate-check (remaining columns are ignored).

    Returns
    -------
    df: pandas.DataFrame
        DataFrame w/o duplicates.
    """
    columns = list(columns)
    dup_mask = mask_sorted_duplicates_keep_last(df, columns)
    if dup_mask.any():
        # pandas is just slow, so try to avoid the indexing call
        return df.iloc[~dup_mask]
    else:
        return df


def aggregate_to_lists(df, by, data_col):
    """Do a group-by and collect the results as python lists.

    Roughly equivalent to::

        df = df.groupby(
            by=by,
            as_index=False,
            observed=True,
        )[data_col].agg(lambda series: list(series.values))

    Parameters
    ----------
    df: pandas.DataFrame
        Dataframe.
    by: Iterable[str]
        Group-by columns, might be empty.
    data_col: str
        Column with values to be collected.

    Returns
    -------
    df: pandas.DataFrame
        DataFrame w/ operation applied.
    """
    by = list(by)

    if df.empty:
        return df

    if not by:
        return pd.DataFrame({data_col: pd.Series([list(df[data_col].values)])})

    # sort the DataFrame by `by`-values, so that rows of every group-by group are consecutive
    df = sort_dataframe(df, by)

    # collect the following data for every group:
    # - by-values
    # - list of values in `data_col`
    result_idx_data: list[Any] = [[] for _ in by]
    result_labels: list[list[Any]] = []

    # remember index (aka values in `by`) and list of data values for current group
    group_idx = None  # Tuple[Any, ...]
    group_values: list[Any] = []

    def _store_group():
        """Store current group from `group_idx` and `group_values` intro result
        lists."""
        if group_idx is None:
            # no group exists yet
            return

        for result_idx_part, idx_part in zip(result_idx_data, group_idx, strict=False):
            result_idx_part.append(idx_part)
        result_labels.append(group_values)

    # create iterator over row-tuples, where every tuple contains values of all by-columns
    iterator_idx = zip(*(df[col].values for col in by), strict=False)

    # iterate over all rows in DataFrame and collect groups
    for idx, label in zip(iterator_idx, df[data_col].values, strict=False):
        if (group_idx is None) or (idx != group_idx):
            _store_group()
            group_idx = idx
            group_values = [label]
        else:
            group_values.append(label)

    # store last group
    _store_group()

    # create result DataFrame out of lists
    data = OrderedDict(zip(by, result_idx_data, strict=False))
    data[data_col] = result_labels
    return pd.DataFrame(data)


def merge_dataframes_robust(df1, df2, how):
    """Merge two given DataFrames but also work if there are no columns to join
    on.

    If now shared column between the given DataFrames is found, then the join will be performaned on a single, constant
    column.

    Parameters
    ----------
    df1: pd.DataFrame
        Left DataFrame.
    df2: pd.DataFrame
        Right DataFrame.
    how: str
        How to join the frames.

    Returns
    -------
    df_joined: pd.DataFrame
        Joined DataFrame.
    """
    dummy_column = "__ktk_join_dummy"

    columns2 = set(df2.columns)
    joined_columns = [c for c in df1.columns if c in columns2]

    if len(joined_columns) == 0:
        df1 = df1.copy()
        df2 = df2.copy()
        df1[dummy_column] = 1
        df2[dummy_column] = 1
        joined_columns = [dummy_column]

    df_out = df1.merge(df2, on=joined_columns, how=how, sort=False)

    df_out.drop(columns=dummy_column, inplace=True, errors="ignore")
    return df_out



import pickle
from functools import partial
from typing import Any, cast

from minimalkv import KeyValueStore, get_store_from_url

from plateau.core.naming import MAX_METADATA_VERSION, MIN_METADATA_VERSION
from plateau.core.typing import StoreFactory, StoreInput

__all__ = ("ensure_store", "lazy_store")


def _verify_metadata_version(metadata_version: int) -> None:
    """This is factored out to be an easier target for mocking."""
    if metadata_version < MIN_METADATA_VERSION:
        raise NotImplementedError(
            f"Minimal supported metadata version is 4. You requested {metadata_version} instead."
        )
    elif metadata_version > MAX_METADATA_VERSION:
        raise NotImplementedError(
            f"Future metadata version `{metadata_version}` encountered."
        )


verify_metadata_version = _verify_metadata_version


def ensure_string_type(obj: bytes | str) -> str:
    """Parse object passed to the function to `str`.

    If the object is of type `bytes`, it is decoded, otherwise a generic string representation of the object is
    returned.

    Parameters
    ----------
    obj
        object which is to be parsed to `str`
    """
    if isinstance(obj, bytes):
        return obj.decode()
    else:
        return str(obj)


def _is_minimalkv_key_value_store(obj: Any) -> bool:
    """Check whether ``obj`` is the ``minimalkv.KeyValueStore``-like class.

    minimalkv uses duck-typing, e.g. for decorators. Therefore, avoid
    `isinstance(store, KeyValueStore)`, as it would be unreliable.
    Instead, only roughly verify that `store` looks like a
    KeyValueStore.
    """
    return hasattr(obj, "iter_prefixes")


def ensure_store(store: StoreInput) -> KeyValueStore:
    """Convert the ``store`` argument to a ``KeyValueStore``, without pickle
    test."""
    # This function is often used in an eager context where we may allow
    # non-serializable stores, so skip the pickle test.
    if _is_minimalkv_key_value_store(store):
        return cast(KeyValueStore, store)
    return lazy_store(store)()


def _identity(store: KeyValueStore) -> KeyValueStore:
    """Helper function for `lazy_store`."""
    return store


def lazy_store(store: StoreInput) -> StoreFactory:
    """Create a store factory from the input. Acceptable inputs are.

    * Storefact store url
    * Callable[[], KeyValueStore]
    * KeyValueStore

    If a KeyValueStore is provided, it is verified that the store is serializable
    (i.e. that pickle.dumps does not raise an exception).
    """
    if callable(store):
        return cast(StoreFactory, store)
    elif isinstance(store, str):
        ret_val = partial(get_store_from_url, store)
        ret_val = cast(StoreFactory, ret_val)  # type: ignore
        return ret_val
    else:
        if not _is_minimalkv_key_value_store(store):
            raise TypeError(
                f"Provided incompatible store type. Got {type(store)} but expected {StoreInput}."
            )

        try:
            pickle.dumps(store, pickle.HIGHEST_PROTOCOL)
        except Exception as exc:
            raise TypeError(
                """KeyValueStore not serializable.
Please consult https://plateau.readthedocs.io/en/stable/spec/store_interface.html for more information."""
            ) from exc
        return partial(_identity, store)



import copy
import logging
import re
from collections import OrderedDict, defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    TypeVar,
    cast,
)

import pandas as pd
import pyarrow as pa
import simplejson

import plateau.core._time
from plateau.core import naming
from plateau.core._compat import load_json
from plateau.core._mixins import CopyMixin
from plateau.core._zmsgpack import packb, unpackb
from plateau.core.common_metadata import SchemaWrapper, read_schema_metadata
from plateau.core.docs import default_docs
from plateau.core.index import (
    ExplicitSecondaryIndex,
    IndexBase,
    PartitionIndex,
    filter_indices,
)
from plateau.core.naming import EXTERNAL_INDEX_SUFFIX, PARQUET_FILE_SUFFIX, SINGLE_TABLE
from plateau.core.partition import Partition
from plateau.core.typing import StoreInput
from plateau.core.urlencode import decode_key, quote_indices
from plateau.core.utils import ensure_store, verify_metadata_version
from plateau.serialization import PredicatesType, columns_in_predicates

if TYPE_CHECKING:
    from minimalkv import KeyValueStore

_logger = logging.getLogger(__name__)

TableMetaType = dict[str, SchemaWrapper]

__all__ = ("DatasetMetadata", "DatasetMetadataBase")


def _validate_uuid(uuid: str) -> bool:
    return re.match(r"[a-zA-Z0-9+\-_]+$", uuid) is not None


def to_ordinary_dict(dct: dict) -> dict:
    new_dct = {}
    for key, value in dct.items():
        if isinstance(value, dict):
            new_dct[key] = to_ordinary_dict(value)
        else:
            new_dct[key] = value
    return new_dct


T = TypeVar("T", bound="DatasetMetadataBase")


class DatasetMetadataBase(CopyMixin):
    def __init__(
        self,
        uuid: str,
        partitions: dict[str, Partition] | None = None,
        metadata: dict | None = None,
        indices: dict[str, IndexBase] | None = None,
        metadata_version: int = naming.DEFAULT_METADATA_VERSION,
        explicit_partitions: bool = True,
        partition_keys: list[str] | None = None,
        schema: SchemaWrapper | None = None,
        table_name: str | None = SINGLE_TABLE,
    ):
        if not _validate_uuid(uuid):
            raise ValueError("UUID contains illegal character")
        self.metadata_version = metadata_version
        self.uuid = uuid
        self.partitions = partitions if partitions else {}
        self.metadata = metadata if metadata else {}
        self.indices = indices if indices else {}
        # explicit partitions means that the partitions are defined in the
        # metadata.json file (in contrast to implicit partitions that are
        # derived from the partition key names)
        self.explicit_partitions = explicit_partitions

        self.partition_keys = partition_keys or []
        self.schema = schema
        self._table_name = table_name

        _add_creation_time(self)
        super().__init__()

    def __eq__(self, other: Any) -> bool:
        # Enforce dict comparison at the places where we only
        # care about content, not order.
        if self.uuid != other.uuid:
            return False
        if to_ordinary_dict(self.partitions) != to_ordinary_dict(other.partitions):
            return False
        if to_ordinary_dict(self.metadata) != to_ordinary_dict(other.metadata):
            return False
        if self.indices != other.indices:
            return False
        if self.explicit_partitions != other.explicit_partitions:
            return False
        if self.partition_keys != other.partition_keys:
            return False
        if self.schema != other.schema:
            return False
        return True

    @property
    def primary_indices_loaded(self) -> bool:
        if not self.partition_keys:
            return False
        for pkey in self.partition_keys:
            if pkey not in self.indices:
                return False
        return True

    @property
    def table_name(self) -> str:
        if self._table_name:
            return self._table_name
        elif self.partitions:
            tables = self.tables
            if tables:
                return self.tables[0]
        return "<Unknown Table>"

    @property
    def tables(self) -> list[str]:
        tables = list(iter(next(iter(self.partitions.values())).files.keys()))
        if len(tables) > 1:
            raise RuntimeError(
                f"Dataset {self.uuid} has tables {tables} but read support for multi tabled dataset was dropped with plateau 4.0."
            )
        return tables

    @property
    def index_columns(self) -> set[str]:
        return set(self.indices.keys()).union(self.partition_keys)

    @property
    def secondary_indices(self) -> dict[str, ExplicitSecondaryIndex]:
        return {
            col: ind
            for col, ind in self.indices.items()
            if isinstance(ind, ExplicitSecondaryIndex)
        }

    @staticmethod
    def exists(uuid: str, store: StoreInput) -> bool:
        """Check if  a dataset exists in a storage.

        Parameters
        ----------
        uuid
            UUID of the dataset.
        store
            Object that implements the .get method for file/object loading.
        """
        store = ensure_store(store)
        key = naming.metadata_key_from_uuid(uuid)

        if key in store:
            return True

        key = naming.metadata_key_from_uuid(uuid, format="msgpack")
        return key in store

    @staticmethod
    def storage_keys(uuid: str, store: StoreInput) -> list[str]:
        """Retrieve all keys that belong to the given dataset.

        Parameters
        ----------
        uuid
            UUID of the dataset.
        store
            Object that implements the .iter_keys method for key retrieval loading.
        """
        store = ensure_store(store)
        start_markers = [f"{uuid}.", f"{uuid}/"]
        return sorted(
            k
            for k in store.iter_keys(uuid)
            if any(k.startswith(marker) for marker in start_markers)
        )

    def to_dict(self) -> dict:
        dct = OrderedDict(
            [
                (naming.METADATA_VERSION_KEY, self.metadata_version),
                (naming.UUID_KEY, self.uuid),
            ]
        )
        if self.indices:
            dct["indices"] = {
                k: v.to_dict()
                if v.loaded
                else v.index_storage_key
                if isinstance(v, ExplicitSecondaryIndex)
                else {}
                for k, v in self.indices.items()
            }
        if self.metadata:
            dct["metadata"] = self.metadata
        if self.partitions or self.explicit_partitions:
            dct["partitions"] = {
                label: partition.to_dict()
                for label, partition in self.partitions.items()
            }

        if self.partition_keys is not None:
            dct["partition_keys"] = self.partition_keys

        return dct

    def to_json(self) -> bytes:
        return simplejson.dumps(self.to_dict()).encode("utf-8")

    def to_msgpack(self) -> bytes:
        return packb(self.to_dict())

    def load_partition_indices(self: T) -> T:
        """Load all filename encoded indices into RAM. File encoded indices can
        be extracted from datasets with partitions stored in a format like.

        .. code::

            `dataset_uuid/table/IndexCol=IndexValue/SecondIndexCol=Value/partition_label.parquet`

        Which results in an in-memory index holding the information

        .. code::

            {
                "IndexCol": {
                    IndexValue: ["partition_label"]
                },
                "SecondIndexCol": {
                    Value: ["partition_label"]
                }
            }
        """
        if self.primary_indices_loaded:
            return self

        indices = _construct_dynamic_index_from_partitions(
            partitions=self.partitions,
            schema=self.schema,
            default_dtype=pa.string() if self.metadata_version == 3 else None,
            partition_keys=self.partition_keys,
        )
        combined_indices = self.indices.copy()
        combined_indices.update(indices)
        return self.copy(indices=combined_indices)

    def load_index(self: T, column: str, store: StoreInput) -> T:
        """Load an index into memory.

        Note: External indices need to be preloaded before they can be queried.

        Parameters
        ----------
        column
            Name of the column for which the index should be loaded.
        store
            Object that implements the .get method for file/object loading.

        Returns
        -------
        dataset_metadata: :class:`~plateau.core.dataset.DatasetMetadata`
            Mutated metadata object with the loaded index.
        """
        if self.partition_keys and column in self.partition_keys:
            return self.load_partition_indices()

        if column not in self.indices:
            raise KeyError(f"No index specified for column '{column}'")

        index = self.indices[column]
        if index.loaded or not isinstance(index, ExplicitSecondaryIndex):
            return self

        loaded_index = index.load(store=store)
        if not self.explicit_partitions:
            col_loaded_index = filter_indices(
                {column: loaded_index}, self.partitions.keys()
            )
        else:
            col_loaded_index = {column: loaded_index}
        indices = dict(self.indices, **col_loaded_index)
        return self.copy(indices=indices)

    def load_all_indices(self: T, store: StoreInput) -> T:
        """Load all registered indices into memory.

        Note: External indices need to be preloaded before they can be queried.

        Parameters
        ----------
        store
            Object that implements the .get method for file/object loading.

        Returns
        -------
        dataset_metadata: :class:`~plateau.core.dataset.DatasetMetadata`
            Mutated metadata object with the loaded indices.
        """
        indices = {
            column: index.load(store)
            if isinstance(index, ExplicitSecondaryIndex)
            else index
            for column, index in self.indices.items()
        }
        ds = self.copy(indices=indices)

        return ds.load_partition_indices()

    def query(self, indices: list[IndexBase] | None = None, **kwargs) -> list[str]:
        """Query the dataset for partitions that contain specific values.
        Lookup is performed using the embedded and loaded external indices.
        Additional indices need to operate on the same partitions that the
        dataset contains, otherwise an empty list will be returned (the query
        method only restricts the set of partition keys using the indices).

        Parameters
        ----------
        indices:
            List of optional additional indices.
        **kwargs:
            Map of columns and values.

        Returns
        -------
        List[str]
            List of keys of partitions that contain the queries values in the respective columns.
        """
        candidate_set = set(self.partitions.keys())

        additional_indices = indices if indices else {}
        combined_indices = dict(
            self.indices, **{index.column: index for index in additional_indices}
        )

        for column, value in kwargs.items():
            if column in combined_indices:
                candidate_set &= set(combined_indices[column].query(value))

        return list(candidate_set)

    @default_docs
    def get_indices_as_dataframe(
        self,
        columns: list[str] | None = None,
        date_as_object: bool = True,
        predicates: PredicatesType = None,
    ):
        """Converts the dataset indices to a pandas dataframe and filter
        relevant indices by `predicates`.

        For a dataset with indices on columns `column_a` and `column_b` and three partitions,
        the dataset output may look like

        .. code::

                    column_a column_b
            part_1         1        A
            part_2         2        B
            part_3         3     None

        Parameters
        ----------
        """
        if self.partition_keys and (
            columns is None
            or (
                self.partition_keys is not None
                and set(columns) & set(self.partition_keys)
            )
        ):
            # self.load_partition_indices is not inplace
            dm = self.load_partition_indices()
        else:
            dm = self

        if columns is None:
            columns = sorted(dm.indices.keys())

        if columns == []:
            return pd.DataFrame(index=dm.partitions)

        if predicates:
            predicate_columns = columns_in_predicates(predicates)
            columns_to_scan = sorted(
                (predicate_columns & self.indices.keys()) | set(columns)
            )

            dfs = (
                dm._evaluate_conjunction(
                    columns=columns_to_scan,
                    predicates=[conjunction],
                    date_as_object=date_as_object,
                )
                for conjunction in predicates
            )

            df = pd.concat(dfs)
            index_name = df.index.name
            df = (
                df.loc[:, columns].reset_index().drop_duplicates().set_index(index_name)
            )
        else:
            df = dm._evaluate_conjunction(
                columns=columns,
                predicates=None,
                date_as_object=date_as_object,
            )
        return df

    def _evaluate_conjunction(
        self, columns: list[str], predicates: PredicatesType, date_as_object: bool
    ) -> pd.DataFrame:
        """Evaluate all predicates related to `columns` to "AND".

        Parameters
        ----------
        columns:
            A list of all columns, including query and index columns.
        predicates:
            Optional list of predicates, like [[('x', '>', 0), ...], that are used
            to filter the resulting DataFrame, possibly using predicate pushdown,
            if supported by the file format.
            This parameter is not compatible with filter_query.

            Predicates are expressed in disjunctive normal form (DNF). This means
            that the innermost tuple describes a single column predicate. These
            inner predicates are all combined with a conjunction (AND) into a
            larger predicate. The most outer list then combines all predicates
            with a disjunction (OR). By this, we should be able to express all
            kinds of predicates that are possible using boolean logic.

            Available operators are: `==`, `!=`, `<=`, `>=`, `<`, `>` and `in`.
        dates_as_object: bool
            Load pyarrow.date{32,64} columns as ``object`` columns in Pandas
            instead of using ``np.datetime64`` to preserve their type. While
            this improves type-safety, this comes at a performance cost.

        Returns
        -------
        pd.DataFrame: df_result
            A DataFrame containing all indices for which `predicates` holds true.
        """
        non_index_columns = set(columns) - self.indices.keys()
        if non_index_columns:
            if non_index_columns & set(self.partition_keys):
                raise RuntimeError(
                    "Partition indices not loaded. Please call `DatasetMetadata.load_partition_indices` first."
                )
            raise ValueError(
                "Unknown index columns: {}".format(", ".join(sorted(non_index_columns)))
            )
        dfs = []
        for col in columns:
            df = pd.DataFrame(
                self.indices[col].as_flat_series(
                    partitions_as_index=True,
                    date_as_object=date_as_object,
                    predicates=predicates,
                )
            )
            dfs.append(df)

        # dfs contains one df per index column. Each df stores indices filtered by `predicates` for each column.
        # Performing an inner join on these dfs yields the resulting "AND" evaluation for all of these predicates.
        # We start joining with the smallest dataframe, therefore the sorting.
        dfs_sorted = sorted(dfs, key=len)
        df_result = dfs_sorted.pop(0)
        for df in dfs_sorted:
            df_result = df_result.merge(
                df, left_index=True, right_index=True, copy=False
            )
        # Backward-compatibility: Set dtype to empty if df is empty.
        # With pandas 2.3, the dtype of the numpy is correctly propagated but changes our API.
        if len(df_result) == 0:
            df_result.index = df_result.index.astype("object")
        return df_result


class DatasetMetadata(DatasetMetadataBase):
    """Containing holding all metadata of the dataset."""

    def __repr__(self):
        return (
            f"DatasetMetadata(uuid={self.uuid}, "
            f"table_name={self.table_name}, "
            f"partition_keys={self.partition_keys}, "
            f"metadata_version={self.metadata_version}, "
            f"indices={list(self.indices.keys())}, "
            f"explicit_partitions={self.explicit_partitions})"
        )

    @staticmethod
    def load_from_buffer(
        buf, store: "KeyValueStore", format: str = "json"
    ) -> "DatasetMetadata":
        """Load a dataset from a (string) buffer.

        Parameters
        ----------
        buf:
            Input to be parsed.
        store:
            Object that implements the .get method for file/object loading.

        Returns
        -------
        DatasetMetadata:
            Parsed metadata.
        """
        if format == "json":
            metadata = load_json(buf)
        elif format == "msgpack":
            metadata = unpackb(buf)
        return DatasetMetadata.load_from_dict(metadata, store)

    @staticmethod
    def load_from_store(
        uuid: str,
        store: StoreInput,
        load_schema: bool = True,
        load_all_indices: bool = False,
    ) -> "DatasetMetadata":
        """Load a dataset from a storage.

        Parameters
        ----------
        uuid
            UUID of the dataset.
        store
            Object that implements the .get method for file/object loading.
        load_schema
            Load table schema
        load_all_indices
            Load all registered indices into memory.

        Returns
        -------
        dataset_metadata: :class:`~plateau.core.dataset.DatasetMetadata`
            Parsed metadata.
        """
        key1 = naming.metadata_key_from_uuid(uuid)
        store = ensure_store(store)
        try:
            value = store.get(key1)
            metadata = load_json(value)
        except KeyError:
            key2 = naming.metadata_key_from_uuid(uuid, format="msgpack")
            try:
                value = store.get(key2)
                metadata = unpackb(value)
            except KeyError as e:
                raise KeyError(
                    f"Dataset does not exist. Tried {key1} and {key2}"
                ) from e

        ds = DatasetMetadata.load_from_dict(metadata, store, load_schema=load_schema)
        if load_all_indices:
            ds = ds.load_all_indices(store)
        return ds

    @staticmethod
    def load_from_dict(
        dct: dict, store: "KeyValueStore", load_schema: bool = True
    ) -> "DatasetMetadata":
        """Load dataset metadata from a dictionary and resolve any external
        includes.

        Parameters
        ----------
        dct
        store
            Object that implements the .get method for file/object loading.
        load_schema
            Load table schema
        """
        # Use copy here to get an OrderedDict
        metadata = copy.copy(dct)

        if "metadata" not in metadata:
            metadata["metadata"] = OrderedDict()

        metadata_version = dct[naming.METADATA_VERSION_KEY]
        dataset_uuid = dct[naming.UUID_KEY]
        explicit_partitions = "partitions" in metadata
        storage_keys = None
        if not explicit_partitions:
            storage_keys = DatasetMetadata.storage_keys(dataset_uuid, store)
            partitions = _load_partitions_from_filenames(
                store=store,
                storage_keys=storage_keys,
                metadata_version=metadata_version,
            )
            metadata["partitions"] = partitions

        if metadata["partitions"]:
            tables = list(list(metadata["partitions"].values())[0]["files"])
        else:
            table_set = set()
            if storage_keys is None:
                storage_keys = DatasetMetadata.storage_keys(dataset_uuid, store)
            for key in storage_keys:
                if key.endswith(naming.TABLE_METADATA_FILE):
                    table_set.add(key.split("/")[1])
            tables = list(table_set)

        schema = None
        table_name = None
        if tables:
            table_name = tables[0]

            if load_schema:
                schema = read_schema_metadata(
                    dataset_uuid=dataset_uuid, store=store, table=table_name
                )

        metadata["schema"] = schema

        if "partition_keys" not in metadata:
            metadata["partition_keys"] = _get_partition_keys_from_partitions(
                metadata["partitions"]
            )

        ds = DatasetMetadata.from_dict(
            metadata, explicit_partitions=explicit_partitions
        )
        if table_name:
            ds._table_name = table_name
        return ds

    @staticmethod
    def from_buffer(buf: str, format: str = "json", explicit_partitions: bool = True):
        if format == "json":
            metadata = load_json(buf)
        else:
            metadata = unpackb(buf)
        return DatasetMetadata.from_dict(
            metadata, explicit_partitions=explicit_partitions
        )

    @staticmethod
    def from_dict(dct: dict, explicit_partitions: bool = True):
        """Load dataset metadata from a dictionary.

        This must have no external references. Otherwise use
        ``load_from_dict`` to have them resolved automatically.
        """

        # Use the builder class for reconstruction to have a single point for metadata version changes
        builder = DatasetMetadataBuilder(
            uuid=dct[naming.UUID_KEY],
            metadata_version=dct[naming.METADATA_VERSION_KEY],
            explicit_partitions=explicit_partitions,
            partition_keys=dct.get("partition_keys", None),
            schema=dct.get("schema"),
        )

        for key, value in dct.get("metadata", {}).items():
            builder.add_metadata(key, value)
        for partition_label, part_dct in dct.get("partitions", {}).items():
            builder.add_partition(
                partition_label, Partition.from_dict(partition_label, part_dct)
            )
        for column, index_dct in dct.get("indices", {}).items():
            if isinstance(index_dct, IndexBase):
                builder.add_embedded_index(column, index_dct)
            else:
                builder.add_embedded_index(
                    column, ExplicitSecondaryIndex.from_v2(column, index_dct)
                )
        return builder.to_dataset()


def _get_type_from_meta(
    schema: SchemaWrapper | None,
    column: str,
    default: pa.DataType | None,
) -> pa.DataType:
    # use first schema that provides type information, since write path should ensure that types are normalized and
    # equal
    if schema is not None:
        idx = schema.get_field_index(column)
        return schema[idx].type

    if default is not None:
        return default

    raise ValueError(f'Cannot find type information for partition column "{column}"')


def _empty_partition_indices(
    partition_keys: list[str],
    schema: SchemaWrapper | None,
    default_dtype: pa.DataType,
):
    indices = {}
    for col in partition_keys:
        arrow_type = _get_type_from_meta(schema, col, default_dtype)
        indices[col] = PartitionIndex(column=col, index_dct={}, dtype=arrow_type)
    return indices


def _construct_dynamic_index_from_partitions(
    partitions: dict[str, Partition],
    schema: SchemaWrapper | None,
    default_dtype: pa.DataType,
    partition_keys: list[str],
) -> dict[str, PartitionIndex]:
    if len(partitions) == 0:
        return _empty_partition_indices(partition_keys, schema, default_dtype)

    def _get_files(part):
        if isinstance(part, dict):
            return part["files"]
        else:
            return part.files

    # We exploit the fact that all tables are partitioned equally.
    first_partition = next(
        iter(partitions.values())
    )  # partitions is NOT empty here, see check above
    first_partition_files = _get_files(first_partition)
    if not first_partition_files:
        return _empty_partition_indices(partition_keys, schema, default_dtype)
    key_table = next(iter(first_partition_files.keys()))
    storage_keys = (
        (key, _get_files(part)[key_table]) for key, part in partitions.items()
    )

    _key_indices: dict[str, dict[str, set[str]]] = defaultdict(_get_empty_index)
    depth_indices = None
    for partition_label, key in storage_keys:
        _, _, indices, file_ = decode_key(key)
        if (
            file_ is not None
            and key.endswith(PARQUET_FILE_SUFFIX)
            and not key.endswith(EXTERNAL_INDEX_SUFFIX)
        ):
            depth_indices = _check_index_depth(indices, depth_indices)
            for column, value in indices:
                _key_indices[column][value].add(partition_label)
    new_indices = {}
    for col, index_dct in _key_indices.items():
        arrow_type = _get_type_from_meta(schema, col, default_dtype)

        # convert defaultdicts into dicts with deterministically ordered values
        new_indices[col] = PartitionIndex(
            column=col,
            index_dct={k1: sorted(v1) for k1, v1 in index_dct.items()},
            dtype=arrow_type,
        )
    return new_indices


def _get_partition_label(indices, filename, metadata_version):
    return "/".join(
        quote_indices(indices) + [filename.replace(PARQUET_FILE_SUFFIX, "")]
    )


def _check_index_depth(indices, depth_indices):
    if depth_indices is not None and len(indices) != depth_indices:
        raise RuntimeError(
            "Unknown file structure encountered. "
            "Depth of filename indices is not equal for all partitions."
        )
    return len(indices)


def _get_partition_keys_from_partitions(partitions):
    if len(partitions):
        part = next(iter(partitions.values()))
        files_dct = part["files"]
        if files_dct:
            key = next(iter(files_dct.values()))
            _, _, indices, _ = decode_key(key)
            if indices:
                return [tup[0] for tup in indices]
    return None


def _load_partitions_from_filenames(store, storage_keys, metadata_version):
    partitions: dict[str, dict[str, Any]] = defaultdict(_get_empty_partition)
    depth_indices = None
    for key in storage_keys:
        dataset_uuid, table, indices, file_ = decode_key(key)
        if file_ is not None and file_.endswith(PARQUET_FILE_SUFFIX):
            # valid key example:
            # <uuid>/<table>/<column_0>=<value_0>/.../<column_n>=<value_n>/part_label.parquet
            depth_indices = _check_index_depth(indices, depth_indices)
            partition_label = _get_partition_label(indices, file_, metadata_version)
            partitions[partition_label]["files"][table] = key
    return partitions


def _get_empty_partition():
    return {"files": {}, "metadata": {}}


def _get_empty_index():
    return defaultdict(set)


def create_partition_key(
    dataset_uuid: str,
    table: str,
    index_values: list[tuple[str, str]],
    filename: str = "data",
):
    """Create partition key for a plateau partition.

    Parameters
    ----------
    dataset_uuid
    table
    index_values
    filename

    Example:
        create_partition_key('my-uuid', 'testtable',
            [('index1', 'value1'), ('index2', 'value2')])

        returns 'my-uuid/testtable/index1=value1/index2=value2/data'
    """
    key_components = [dataset_uuid, table]
    index_path = quote_indices(index_values)
    key_components.extend(index_path)
    key_components.append(filename)
    key = "/".join(key_components)
    return key


class DatasetMetadataBuilder(CopyMixin):
    """Incrementally build up a dataset.

    In contrast to a :class:`plateau.core.dataset.DatasetMetadata`
    instance, this object is mutable and may not be a full dataset (e.g.
    partitions don't need to be fully materialised).
    """

    def __init__(
        self,
        uuid: str,
        metadata_version=naming.DEFAULT_METADATA_VERSION,
        explicit_partitions=True,
        partition_keys=None,
        schema=None,
    ):
        verify_metadata_version(metadata_version)

        self.uuid = uuid
        self.metadata: dict = OrderedDict()
        self.indices: dict[str, IndexBase] = OrderedDict()
        self.metadata_version = metadata_version
        self.partitions: dict[str, Partition] = OrderedDict()
        self.partition_keys = partition_keys
        self.schema = schema
        self.explicit_partitions = explicit_partitions

        _add_creation_time(self)
        super().__init__()

    @staticmethod
    def from_dataset(dataset):
        dataset = copy.deepcopy(dataset)

        ds_builder = DatasetMetadataBuilder(
            uuid=dataset.uuid,
            metadata_version=dataset.metadata_version,
            explicit_partitions=dataset.explicit_partitions,
            partition_keys=dataset.partition_keys,
            schema=dataset.schema,
        )

        ds_builder.metadata = dataset.metadata
        ds_builder.indices = dataset.indices
        ds_builder.partitions = dataset.partitions
        return ds_builder

    def add_partition(self, name, partition):
        """Add an (embedded) Partition.

        Parameters
        ----------
        name: str
            Identifier of the partition.
        partition: :class:`plateau.core.partition.Partition`
            The partition to add.
        """

        if len(partition.files) > 1:
            raise RuntimeError(
                f"Dataset {self.uuid} has tables {sorted(partition.files.keys())} but read support for multi tabled dataset was dropped with plateau 4.0."
            )

        self.partitions[name] = partition
        return self

    # TODO: maybe remove
    def add_embedded_index(self, column, index):
        """Embed an index into the metadata.

        Parameters
        ----------
        column: str
            Name of the indexed column
        index: plateau.core.index.IndexBase
            The actual index object
        """

        if column != index.column:
            # TODO Deprecate the column argument and take the column name directly from the index.
            raise RuntimeError(
                "The supplied index is not compatible with the supplied index."
            )

        self.indices[column] = index

    def add_external_index(self, column, filename=None):
        """Add a reference to an external index.

        Parameters
        ----------
        column: str
            Name of the indexed column

        Returns
        -------
        storage_key: str
            The location where the external index should be stored.
        """
        if filename is None:
            filename = f"{self.uuid}.{column}"
            filename += naming.EXTERNAL_INDEX_SUFFIX
        self.indices[column] = ExplicitSecondaryIndex(
            column, index_storage_key=filename
        )
        return filename

    def add_metadata(self, key, value):
        """Add arbitrary key->value metadata.

        Parameters
        ----------
        key: str
        value: str
        """
        self.metadata[key] = value

    def to_dict(self):
        """Render the dataset to a dict.

        Returns
        -------
        """
        factory = type(self.metadata)
        dct = factory(
            [
                (naming.METADATA_VERSION_KEY, self.metadata_version),
                (naming.UUID_KEY, self.uuid),
            ]
        )
        if self.indices:
            dct["indices"] = {}
            for column, index in self.indices.items():
                if isinstance(index, str):
                    dct["indices"][column] = index
                elif index.loaded:
                    dct["indices"][column] = index.to_dict()
                else:
                    dct["indices"][column] = cast(
                        ExplicitSecondaryIndex, index
                    ).index_storage_key
        if self.metadata:
            dct["metadata"] = self.metadata

        if self.explicit_partitions:
            dct["partitions"] = factory()
            for label, partition in self.partitions.items():
                part_dict = partition.to_dict()
                dct["partitions"][label] = part_dict

        if self.partition_keys is not None:
            dct["partition_keys"] = self.partition_keys
        return dct

    def to_json(self):
        """Render the dataset to JSON.

        Returns
        -------
        storage_key: str
            The path where this metadata should be placed in the storage.
        dataset_json: str
            The rendered JSON for this dataset.
        """
        return (
            naming.metadata_key_from_uuid(self.uuid),
            simplejson.dumps(self.to_dict()).encode("utf-8"),
        )

    def to_msgpack(self) -> tuple[str, bytes]:
        """Render the dataset to msgpack.

        Returns
        -------
        storage_key: str
            The path where this metadata should be placed in the storage.
        dataset_json: str
            The rendered JSON for this dataset.
        """
        return (
            naming.metadata_key_from_uuid(self.uuid, format="msgpack"),
            packb(self.to_dict()),
        )

    def to_dataset(self) -> DatasetMetadata:
        return DatasetMetadata(
            uuid=self.uuid,
            partitions=self.partitions,
            metadata=self.metadata,
            indices=self.indices,
            metadata_version=self.metadata_version,
            explicit_partitions=self.explicit_partitions,
            partition_keys=self.partition_keys,
            schema=self.schema,
        )


def _add_creation_time(
    dataset_object: DatasetMetadataBase | DatasetMetadataBuilder,
):
    if "creation_time" not in dataset_object.metadata:
        creation_time = plateau.core._time.datetime_utcnow().isoformat()
        dataset_object.metadata["creation_time"] = creation_time



from copy import deepcopy

from plateau.core.common_metadata import (
    make_meta,
    validate_compatible,
    validate_shared_columns,
)
from plateau.core.testing import get_dataframe_alltypes

from .config import AsvBenchmarkConfig


class TimeMakeMeta(AsvBenchmarkConfig):
    def setup(self):
        self.df = get_dataframe_alltypes()

    def time_make_meta(self):
        make_meta(self.df, origin="df")


class TimeValidateCompatible(AsvBenchmarkConfig):
    params = ([2, 10**2, 10**3, 10**4], [True, False])
    timeout = 120.0

    param_names = ["num_schemas", "has_na"]

    def setup(self, num_schemas, has_na):
        self.df = get_dataframe_alltypes()
        schema = make_meta(self.df, origin="df")
        self.schemas = [deepcopy(schema) for _ in range(num_schemas)]
        if has_na:
            empty_schema = make_meta(self.df[0:0], origin="empty")
            # insert alternating empty schemas
            self.schemas[::2] = [
                deepcopy(empty_schema) for _ in range(len(self.schemas[::2]))
            ]

    def mem_schemas(self, num_schemas, has_na):
        return self.schemas

    def peakmem_validate_compatible(self, num_schemas, has_na):
        validate_compatible(self.schemas)

    def time_validate_compatible(self, num_schemas, has_na):
        validate_compatible(self.schemas)


class TimeValidateSharedColumns(AsvBenchmarkConfig):
    params = [2, 10**2]
    timeout = 120.0

    param_names = ["num_schemas"]

    def setup(self, num_schemas):
        self.df = get_dataframe_alltypes()
        schema = make_meta(self.df, origin="df")
        self.schemas = [deepcopy(schema) for _ in range(num_schemas)]

    def peakmem_validate_shared_columns(self, num_schemas):
        validate_shared_columns(self.schemas)

    def time_validate_shared_columns(self, num_schemas):
        validate_shared_columns(self.schemas)



"""Global naming constants for datasets."""

# FIXME: move this constant somewhere else. Cannot import from its declaration due to cyclic imports
SINGLE_TABLE = "table"
DEFAULT_METADATA_VERSION = 4
MIN_METADATA_VERSION = 4
MAX_METADATA_VERSION = 4

DEFAULT_METADATA_STORAGE_FORMAT = "json"

# Format suffixes
METADATA_FORMAT_JSON = ".json"
METADATA_FORMAT_MSGPACK = ".msgpack.zstd"

# Metadata suffixes
METADATA_BASE_SUFFIX = ".by-dataset-metadata"

# Object suffixes
PARQUET_FILE_SUFFIX = ".parquet"
EXTERNAL_INDEX_SUFFIX = f".by-dataset-index{PARQUET_FILE_SUFFIX}"

METADATA_VERSION_KEY = "dataset_metadata_version"
UUID_KEY = "dataset_uuid"

# Files/BLOB with special meaning

TABLE_METADATA_FILE = "_common_metadata"


def metadata_key_from_uuid(uuid, format="json"):
    if format == "json":
        return uuid + METADATA_BASE_SUFFIX + METADATA_FORMAT_JSON
    elif format == "msgpack":
        return uuid + METADATA_BASE_SUFFIX + METADATA_FORMAT_MSGPACK


def get_partition_file_prefix(
    dataset_uuid, partition_label, table, metadata_version=DEFAULT_METADATA_VERSION
):
    if metadata_version == 4:
        file_prefix = f"{dataset_uuid}/{table}/{partition_label}"
    else:
        raise NotImplementedError(f"Metadata version {metadata_version} not supported")
    return file_prefix



import math

import numpy as np
import pandas as pd
from pyarrow.parquet import ParquetFile

from plateau.io.eager import store_dataframes_as_dataset


def create_dataset(dataset_uuid, store_factory, metadata_version):
    df = pd.DataFrame(
        {"P": np.arange(0, 10), "L": np.arange(0, 10), "TARGET": np.arange(10, 20)}
    )

    df_list = [df.copy(deep=True), df.copy(deep=True)]

    return store_dataframes_as_dataset(
        dfs=df_list,
        store=store_factory,
        dataset_uuid=dataset_uuid,
        metadata_version=metadata_version,
        secondary_indices="P",
    )


def assert_num_row_groups(store, dataset, part_num_rows, part_chunk_size):
    """Assert that the row groups of each partition match the expectation based
    on the number of rows and the chunk size."""
    # Iterate over the partitions of each index value
    for index, partitions in dataset.indices["p"].index_dct.items():
        for part_key in partitions:
            key = dataset.partitions[part_key].files["table"]
            parquet_file = ParquetFile(store.open(key))
            if part_chunk_size[index] is None:
                assert parquet_file.num_row_groups == 1
            else:
                assert parquet_file.num_row_groups == math.ceil(
                    part_num_rows[index] / part_chunk_size[index]
                )



import math
import types

import numpy as np
import pandas as pd
import pytest

from plateau.io.eager import store_dataframes_as_dataset
from plateau.io_components.metapartition import SINGLE_TABLE, MetaPartition
from plateau.io_components.read import dispatch_metapartitions


def test_dispatch_metapartitions(dataset, store_session):
    part_generator = dispatch_metapartitions(dataset.uuid, store_session)

    assert isinstance(part_generator, types.GeneratorType)
    partitions = list(part_generator)

    assert len(partitions) == 2
    assert len({mp.label for mp in partitions}) == 2
    for mp in partitions:
        assert isinstance(mp, MetaPartition)
        assert mp.table_name == SINGLE_TABLE


@pytest.mark.parametrize(
    "predicates,error_msg",
    [([], "Empty predicates"), ([[]], "Invalid predicates: Conjunction 0 is empty")],
)
def test_dispatch_metapartition_undefined_behaviour(
    dataset, store_session, predicates, error_msg
):
    with pytest.raises(ValueError, match=error_msg):
        list(
            dispatch_metapartitions(dataset.uuid, store_session, predicates=predicates)
        )


@pytest.mark.parametrize(
    "predicates",
    [
        [[("P", "==", 2)]],
        [[("P", "in", [2])]],
        [[("P", "!=", 1)]],
        [[("P", ">", 1)]],
        # Only apply filter to columns for which we have an index
        [[("P", ">=", 2), ("TARGET", "==", 500)]],
    ],
)
def test_dispatch_metapartitions_query_partition_on(
    dataset_partition_keys, store_session, predicates
):
    generator = dispatch_metapartitions(
        dataset_partition_keys.uuid, store_session, predicates=predicates
    )
    partitions = list(generator)
    assert len(partitions) == 1
    assert partitions[0].label == "P=2/cluster_2"


@pytest.mark.parametrize(
    "predicates",
    [
        # These predicates are OR connected, therefore they need to allow all partitions
        [[("P", "==", 2)], [("TARGET", "==", 500)]],
        [[("P", "in", [2])], [("TARGET", "in", [500])]],
        [[("L", "==", 2)], [("TARGET", "==", 500)]],
    ],
)
def test_dispatch_metapartitions_query_no_effect(
    dataset_partition_keys, store_session, predicates
):
    # These predicates should still lead to loading the whole set of partitions
    generator = dispatch_metapartitions(
        dataset_partition_keys.uuid, store_session, predicates=predicates
    )
    partitions = list(generator)
    assert len(partitions) == 2


def test_dispatch_metapartitions_concat_regression(store):
    dataset = store_dataframes_as_dataset(
        dfs=[pd.DataFrame({"p": [0], "x": [0]}), pd.DataFrame({"p": [0], "x": [1]})],
        dataset_uuid="test",
        store=store,
        partition_on=["p"],
    )

    mps = list(dispatch_metapartitions(dataset.uuid, store))
    assert len(mps) == 2

    mps = list(dispatch_metapartitions(dataset.uuid, store, dispatch_by=["p"]))
    assert len(mps) == 1


def test_dispatch_metapartitions_dups_with_predicates(store):
    dataset = store_dataframes_as_dataset(
        dfs=[pd.DataFrame({"p": [0, 1], "x": 0})],
        dataset_uuid="test",
        store=store,
        secondary_indices=["p"],
    )

    wout_preds = list(dispatch_metapartitions(dataset.uuid, store))
    w_preds = list(
        dispatch_metapartitions(dataset.uuid, store, predicates=[[("p", "in", [0, 1])]])
    )

    assert wout_preds == w_preds


def test_dispatch_metapartitions_dups_with_predicates_dispatch_by(store):
    dataset = store_dataframes_as_dataset(
        dfs=[pd.DataFrame({"p": [0, 1], "x": 0})],
        dataset_uuid="test",
        store=store,
        secondary_indices=["p", "x"],
    )

    wout_preds = list(dispatch_metapartitions(dataset.uuid, store, dispatch_by="x"))
    w_preds = list(
        dispatch_metapartitions(
            dataset.uuid, store, predicates=[[("p", "in", [0, 1])]], dispatch_by="x"
        )
    )

    assert wout_preds == w_preds


def test_dispatch_metapartitions_sorted_dispatch_by(store):
    df = pd.DataFrame(
        {"p": np.random.randint(high=100000, low=-100000, size=(100,)), "x": 0}
    )
    # Integers are sorted when using too small values (maybe connected to the
    # singleton implementation of integers in CPython??)
    # Verify this is not happening, otherwise we'll get immediately a sorted
    # index (which is nice in this case but not generally true, of course)
    arr = set(df["p"].unique())
    assert list(arr) != sorted(arr)

    dataset = store_dataframes_as_dataset(
        dfs=[df], dataset_uuid="test", store=store, secondary_indices=["p", "x"]
    )

    wout_preds = list(dispatch_metapartitions(dataset.uuid, store, dispatch_by="p"))
    last = -math.inf
    for mps in wout_preds:
        for mp in mps:
            current = mp.logical_conjunction
            assert len(current) == 1
            current = current[0][2]
            assert current > last
            last = current


def test_dispatch_metapartitions_complex_or_predicates(store_factory):
    dataset_uuid = "test"
    df = pd.DataFrame({"A": range(10), "B": ["A", "B"] * 5, "C": range(-10, 0)})

    store_dataframes_as_dataset(
        store=store_factory,
        dataset_uuid=dataset_uuid,
        dfs=[df],
        partition_on=["A", "B"],
    )
    predicates = [[("A", "<", 3)], [("B", "==", "B")]]
    mps = [
        mp.load_dataframes(store_factory)
        for mp in dispatch_metapartitions(
            dataset_uuid, store_factory, predicates=predicates
        )
    ]
    actual = pd.concat([mp.data for mp in mps])
    actual = actual.sort_values(by="A", ignore_index=True).reset_index(drop=True)
    expected = pd.DataFrame(
        data={
            "A": [0, 1, 2, 3, 5, 7, 9],
            "B": ["A", "B", "A", "B", "B", "B", "B"],
            "C": [-10, -9, -8, -7, -5, -3, -1],
        }
    )
    pd.testing.assert_frame_equal(actual, expected)

    predicates = [[("A", "<", 3)], [("B", "==", "notthere")]]
    mps = [
        mp.load_dataframes(store_factory)
        for mp in dispatch_metapartitions(
            dataset_uuid, store_factory, predicates=predicates
        )
    ]
    actual = pd.concat([mp.data for mp in mps])
    actual = actual.sort_values(by="A", ignore_index=True).reset_index(drop=True)
    expected = pd.DataFrame(
        data={"A": [0, 1, 2], "B": ["A", "B", "A"], "C": [-10, -9, -8]}
    )
    pd.testing.assert_frame_equal(actual, expected)

    predicates = [[("A", "<", 3), ("B", "==", "A")], [("B", "==", "B"), ("A", ">", 2)]]
    mps = [
        mp.load_dataframes(store_factory)
        for mp in dispatch_metapartitions(
            dataset_uuid, store_factory, predicates=predicates
        )
    ]
    actual = pd.concat([mp.data for mp in mps])
    actual = actual.sort_values(by="A", ignore_index=True).reset_index(drop=True)
    expected = pd.DataFrame(
        data={
            "A": [0, 2, 3, 5, 7, 9],
            "B": ["A", "A", "B", "B", "B", "B"],
            "C": [-10, -8, -7, -5, -3, -1],
        }
    )
    pd.testing.assert_frame_equal(actual, expected)

    predicates = [[("A", "<", 3)], [("B", "==", "B"), ("A", ">", 2)]]
    mps = [
        mp.load_dataframes(store_factory)
        for mp in dispatch_metapartitions(
            dataset_uuid, store_factory, predicates=predicates
        )
    ]
    actual = pd.concat([mp.data for mp in mps])
    actual = actual.sort_values(by="A", ignore_index=True).reset_index(drop=True)
    expected = pd.DataFrame(
        data={
            "A": [0, 1, 2, 3, 5, 7, 9],
            "B": ["A", "B", "A", "B", "B", "B", "B"],
            "C": [-10, -9, -8, -7, -5, -3, -1],
        }
    )
    pd.testing.assert_frame_equal(actual, expected)



from collections.abc import Iterator
from typing import cast, overload

import pandas as pd

from plateau.core.factory import DatasetFactory
from plateau.core.typing import StoreInput
from plateau.io_components.metapartition import MetaPartition
from plateau.io_components.utils import normalize_args
from plateau.serialization import (
    PredicatesType,
    check_predicates,
    columns_in_predicates,
)


@overload
def dispatch_metapartitions_from_factory(
    dataset_factory: DatasetFactory,
    predicates: PredicatesType = None,
    dispatch_by: None = None,
) -> Iterator[MetaPartition]: ...


@overload
def dispatch_metapartitions_from_factory(
    dataset_factory: DatasetFactory,
    predicates: PredicatesType,
    dispatch_by: list[str],
) -> Iterator[list[MetaPartition]]: ...


@normalize_args
def dispatch_metapartitions_from_factory(
    dataset_factory: DatasetFactory,
    predicates: PredicatesType = None,
    dispatch_by: list[str] | None = None,
) -> Iterator[MetaPartition] | Iterator[list[MetaPartition]]:
    """:meta private:"""

    if dispatch_by is not None and not set(dispatch_by).issubset(
        set(dataset_factory.index_columns)
    ):
        raise RuntimeError(
            f"Dispatch columns must be indexed.\nRequested index: {dispatch_by} but available index columns: {sorted(dataset_factory.index_columns)}"
        )
    check_predicates(predicates)

    # Determine which indices need to be loaded.
    index_cols: set[str] = set()
    if dispatch_by:
        index_cols |= set(dispatch_by)

    if predicates:
        predicate_cols = set(columns_in_predicates(predicates))
        predicate_index_cols = predicate_cols & set(dataset_factory.index_columns)
        index_cols |= predicate_index_cols

    for col in index_cols:
        dataset_factory.load_index(col)

    base_df = dataset_factory.get_indices_as_dataframe(
        list(index_cols), predicates=predicates
    )

    if dispatch_by is not None:
        base_df = cast(pd.DataFrame, base_df)

        if len(dispatch_by) == 0:
            merged_partitions = [((""), base_df)]
        else:
            # Group the resulting MetaPartitions by partition keys or a subset of those keys
            merged_partitions = base_df.groupby(
                by=list(dispatch_by), sort=True, as_index=False, observed=True
            )

        for group_name, group in merged_partitions:
            if not isinstance(group_name, tuple):
                group_name = (group_name,)  # type: ignore
            mps = []
            logical_conjunction = list(
                zip(dispatch_by, ["=="] * len(dispatch_by), group_name, strict=False)
            )
            for label in group.index.unique():
                mps.append(
                    MetaPartition.from_partition(
                        partition=dataset_factory.partitions[label],
                        metadata_version=dataset_factory.metadata_version,
                        schema=dataset_factory.schema,
                        partition_keys=dataset_factory.partition_keys,
                        logical_conjunction=logical_conjunction,
                        table_name=dataset_factory.table_name,
                    )
                )
            yield mps
    else:
        for part_label in base_df.index.unique():
            part = dataset_factory.partitions[part_label]

            yield MetaPartition.from_partition(
                partition=part,
                metadata_version=dataset_factory.metadata_version,
                schema=dataset_factory.schema,
                partition_keys=dataset_factory.partition_keys,
                table_name=dataset_factory.table_name,
            )


def dispatch_metapartitions(
    dataset_uuid: str,
    store: StoreInput,
    predicates: PredicatesType = None,
    dispatch_by: list[str] | None = None,
) -> Iterator[MetaPartition] | Iterator[list[MetaPartition]]:
    dataset_factory = DatasetFactory(
        dataset_uuid=dataset_uuid,
        store_factory=store,
        load_schema=True,
        load_all_indices=False,
    )

    return dispatch_metapartitions_from_factory(
        dataset_factory=dataset_factory,
        predicates=predicates,
        dispatch_by=dispatch_by,
    )



# pylint: disable=E1101


from datetime import date
from functools import partial

import numpy as np
import pandas as pd
import pandas.testing as pdt
import pytest

from plateau.api.dataset import read_dataset_as_ddf
from plateau.core._compat import PANDAS_3
from plateau.core.dataset import DatasetMetadata
from plateau.core.naming import DEFAULT_METADATA_VERSION
from plateau.core.testing import TIME_TO_FREEZE_ISO
from plateau.io.eager import (
    read_dataset_as_metapartitions,
    read_table,
    store_dataframes_as_dataset,
)
from plateau.io.iter import read_dataset_as_dataframes__iterator


def test_update_dataset_with_partitions(
    store_factory, metadata_version, bound_update_dataset, mocker, store
):
    partitions = [
        pd.DataFrame({"p": [1]}),
        pd.DataFrame({"p": [2]}),
    ]
    dataset = bound_update_dataset(
        partitions,
        store=store_factory,
        metadata={"dataset": "metadata"},
        dataset_uuid="dataset_uuid",
        default_metadata_version=metadata_version,
        secondary_indices=["p"],
    )
    dataset = dataset.load_index("p", store)

    dataset_updated = bound_update_dataset(
        [pd.DataFrame({"p": [3]})],
        store=store_factory,
        delete_scope=[{"p": 1}],
        metadata={"extra": "metadata"},
        dataset_uuid="dataset_uuid",
        default_metadata_version=metadata_version,
        secondary_indices=["p"],
    )
    dataset_updated = dataset_updated.load_index("p", store)

    # First check the index. if they are as expected, use them to determine label name

    exp_idx_values = {1, 2}
    ind = dataset.indices["p"]
    ind_updated = dataset_updated.indices["p"]
    assert set(ind.index_dct.keys()) == exp_idx_values
    exp_updated_idx_values = {2, 3}
    assert set(ind_updated.index_dct.keys()) == exp_updated_idx_values

    assert ind_updated.index_dct[2] == ind.index_dct[2]
    assert ind.index_dct[1] != ind_updated.index_dct[3]

    expected_metadata = {"dataset": "metadata", "extra": "metadata"}

    # We do not mock the time resolution of this test since otherwise we cannot ensure
    # that the indices are not overwritten
    expected_metadata["creation_time"] = dataset_updated.metadata["creation_time"]

    assert dataset_updated.metadata == expected_metadata
    assert dataset_updated.uuid == "dataset_uuid"

    store_files = list(store.keys())
    # 1 dataset metadata file and 2 index file and 3 partition files

    # common metadata for v4 datasets (1 table)
    expected_number_files = 7

    assert len(store_files) == expected_number_files

    # Ensure the dataset can be loaded properly
    stored_dataset = DatasetMetadata.load_from_store("dataset_uuid", store)
    stored_dataset = stored_dataset.load_index("p", store)
    assert dataset_updated == stored_dataset


@pytest.mark.xfail(reason="How to handle empty input??")
def test_update_dataset_with_partitions_delete_only(
    store_factory, metadata_version, frozen_time_em, bound_update_dataset, store
):
    partitions = [
        pd.DataFrame({"p": [1]}),
        pd.DataFrame({"p": [2]}),
    ]
    dataset = store_dataframes_as_dataset(
        dfs=partitions,
        store=store_factory,
        metadata={"dataset": "metadata"},
        dataset_uuid="dataset_uuid",
        secondary_indices=["p"],
        metadata_version=metadata_version,
    )
    dataset = dataset.load_index("p", store)

    # FIXME: is this a regression?
    dataset_updated = bound_update_dataset(
        None,
        store=store_factory,
        dataset_uuid="dataset_uuid",
        delete_scope=[{"p": 1}],
        metadata={"extra": "metadata"},
        default_metadata_version=metadata_version,
        secondary_indices=["p"],
    )
    dataset_updated = dataset_updated.load_index("p", store)

    assert len(dataset.partitions) == 2
    assert len(dataset_updated.partitions) == 1

    store_files = list(store.keys())
    # 1 dataset metadata file and 1 index file and 2 partition files
    # note: the update writes a new index file but due to frozen_time this gets
    # the same name as the previous one and overwrites it.
    expected_number_files = 4
    # common metadata for v4 datasets (1 table)
    expected_number_files += 1
    assert len(store_files) == expected_number_files

    assert set(dataset.indices["p"].observed_values()) == {1, 2}
    assert set(dataset_updated.indices["p"].observed_values()) == {2}

    # Ensure the dataset can be loaded properly
    stored_dataset = DatasetMetadata.load_from_store("dataset_uuid", store)
    stored_dataset = stored_dataset.load_index("p", store)
    assert dataset_updated == stored_dataset


def test_update_dataset_with_partitions__reducer_partitions(
    store_factory, frozen_time_em, bound_update_dataset
):
    assert set(store_factory().keys()) == set()

    df1 = pd.DataFrame(
        {"P": [1, 2, 3, 1, 2, 3], "L": [1, 1, 1, 1, 1, 1], "TARGET": np.arange(10, 16)}
    )
    df2 = df1.copy(deep=True)
    df2.L = 2
    df2.TARGET += 2
    df_list = [
        df1,
        df2,
    ]
    dataset = store_dataframes_as_dataset(
        dfs=df_list,
        store=store_factory,
        dataset_uuid="dataset_uuid",
        partition_on=["P"],
        secondary_indices="L",
        metadata_version=4,
    )
    dataset_loadedidx = dataset.load_all_indices(store=store_factory())
    cluster_1_label = (
        dataset_loadedidx.indices["L"].eval_operator(op="==", value=1).pop()
    )
    cluster_1_label = cluster_1_label.split("/")[-1]
    cluster_2_label = (
        dataset_loadedidx.indices["L"].eval_operator(op="==", value=2).pop()
    )
    cluster_2_label = cluster_2_label.split("/")[-1]

    df3 = df2.copy(deep=True)
    df3.TARGET -= 5

    dataset_updated = bound_update_dataset(
        [df3],
        store=store_factory,
        dataset_uuid="dataset_uuid",
        delete_scope=[{"L": 2}],
        metadata={"extra": "metadata"},
        partition_on=["P"],
        secondary_indices=["L"],
    )
    dataset_updated_loadedidx = dataset_updated.load_all_indices(store=store_factory())
    cluster_3_labels = dataset_updated_loadedidx.indices["L"].eval_operator(
        op="==", value=2
    )

    cluster_3_label = {c3_label.split("/")[-1] for c3_label in cluster_3_labels}
    assert len(cluster_3_label) == 1
    cluster_3_label = cluster_3_label.pop()
    exp_partitions = [
        f"P=1/{cluster_1_label}",
        f"P=1/{cluster_3_label}",
        f"P=2/{cluster_1_label}",
        f"P=2/{cluster_3_label}",
        f"P=3/{cluster_1_label}",
        f"P=3/{cluster_3_label}",
    ]
    assert sorted(exp_partitions) == sorted(dataset_updated.partitions.keys())
    updated_idx_keys = sorted(dataset_updated.indices.keys())
    assert sorted(dataset.indices.keys()) == updated_idx_keys

    expected_new_idx = {}
    for k, v in dataset_loadedidx.indices["P"].index_dct.items():
        val = [pl.replace(cluster_2_label, cluster_3_label) for pl in v]
        expected_new_idx[k] = val

    updated_P_idx_dct = dataset_updated_loadedidx.indices["P"].index_dct

    assert sorted(expected_new_idx.keys()) == sorted(updated_P_idx_dct.keys())

    for k, v in updated_P_idx_dct.items():
        assert sorted(expected_new_idx[k]) == sorted(v)


def test_update_dataset_with_partitions__reducer_nonexistent(
    store_factory, metadata_version, frozen_time_em, bound_update_dataset, store
):
    dataset_updated = bound_update_dataset(
        [pd.DataFrame({"p": [3]})],
        store=store_factory,
        dataset_uuid="dataset_uuid",
        delete_scope=[{"p": 1}],
        metadata={"extra": "metadata"},
        default_metadata_version=metadata_version,
        secondary_indices=["p"],
    )
    dataset_updated = dataset_updated.load_index("p", store)
    ind_updated = dataset_updated.indices["p"]
    cluster_3_label = ind_updated.eval_operator(op="==", value=3).pop()

    expected_metadata = {"extra": "metadata"}

    expected_metadata["creation_time"] = TIME_TO_FREEZE_ISO

    assert dataset_updated.metadata == expected_metadata
    assert list(dataset_updated.partitions) == [cluster_3_label]

    updated_part_c3 = dataset_updated.partitions[cluster_3_label]

    assert updated_part_c3.label == cluster_3_label
    assert dataset_updated.uuid == "dataset_uuid"

    store_files = list(store.keys())
    # 1 dataset metadata file and 1 index file and 1 partition files
    # note: the update writes a new index file but due to frozen_time this gets
    # the same name as the previous one and overwrites it.
    expected_number_files = 3

    # common metadata for v4 datasets (1 table)
    expected_number_files += 1
    assert len(store_files) == expected_number_files
    exp_updated_idx = {3: [cluster_3_label]}
    assert dataset_updated.indices["p"].index_dct == exp_updated_idx

    # Ensure the dataset can be loaded properly
    stored_dataset = DatasetMetadata.load_from_store("dataset_uuid", store)
    stored_dataset = stored_dataset.load_index("p", store)
    assert dataset_updated == stored_dataset


@pytest.mark.parametrize(
    "dfs,ok",
    [
        (
            [
                pd.DataFrame(
                    {
                        "P": pd.Series([1], dtype=np.int64),
                        "X": pd.Series([1], dtype=np.int64),
                    }
                ),
                pd.DataFrame(
                    {
                        "P": pd.Series([2], dtype=np.int64),
                        "X": pd.Series([2], dtype=np.int64),
                    }
                ),
            ],
            True,
        ),
        (
            [
                pd.DataFrame(
                    {
                        "P": pd.Series([1], dtype=np.int64),
                        "X": pd.Series([1], dtype=np.int32),
                    }
                ),
                pd.DataFrame(
                    {
                        "P": pd.Series([2], dtype=np.int64),
                        "X": pd.Series([2], dtype=np.int16),
                    }
                ),
            ],
            True,
        ),
        (
            [
                pd.DataFrame(
                    {
                        "P": pd.Series([1], dtype=np.int16),
                        "X": pd.Series([1], dtype=np.int64),
                    }
                ),
                pd.DataFrame(
                    {
                        "P": pd.Series([2], dtype=np.int32),
                        "X": pd.Series([2], dtype=np.int64),
                    }
                ),
            ],
            True,
        ),
        (
            [
                pd.DataFrame(
                    {
                        "P": pd.Series([1], dtype=np.int64),
                        "X": pd.Series([1], dtype=np.int64),
                    }
                ),
                pd.DataFrame(
                    {
                        "P": pd.Series([2], dtype=np.int64),
                        "X": pd.Series([2], dtype=np.uint64),
                    }
                ),
            ],
            False,
        ),
        (
            [
                pd.DataFrame(
                    {
                        "P": pd.Series([1], dtype=np.int64),
                        "X": pd.Series([1], dtype=np.int64),
                    }
                ),
                pd.DataFrame(
                    {
                        "P": pd.Series([2], dtype=np.int64),
                        "X": pd.Series([2], dtype=np.int64),
                        "Y": pd.Series([2], dtype=np.int64),
                    }
                ),
            ],
            False,
        ),
        (
            [
                pd.DataFrame(
                    {
                        "P": pd.Series([1, 2], dtype=np.int64),
                        "X": pd.Series([1, 2], dtype=np.int64),
                    }
                ),
                pd.DataFrame(
                    {
                        "P": pd.Series([3], dtype=np.int64),
                        "X": pd.Series([3], dtype=np.uint64),
                    }
                ),
            ],
            False,
        ),
    ],
)
def test_schema_check_update(dfs, ok, store_factory, bound_update_dataset):
    store_dataframes_as_dataset(
        dfs=dfs[:1],
        store=store_factory,
        dataset_uuid="dataset_uuid",
        partition_on=["P"],
        metadata_version=4,
    )
    pipe = partial(
        bound_update_dataset,
        store=store_factory,
        dataset_uuid="dataset_uuid",
        partition_on=["P"],
    )

    if ok:
        pipe(dfs[1:])
    else:
        with pytest.raises(
            Exception,
            match=r"Schemas\sfor\sdataset\s\\*'dataset_uuid\\*'\sare\snot\scompatible!",
        ):
            pipe(dfs[1:])


def test_sort_partitions_by(
    store_factory, metadata_version, frozen_time_em, bound_update_dataset
):
    df1 = pd.DataFrame({"P": [3], "L": [1], "TARGET": [1]})
    df2 = pd.DataFrame(
        {
            "P": [1, 2, 3, 1, 2, 3],
            "L": [1, 1, 1, 1, 1, 1],
            "TARGET": list(reversed(np.arange(10, 16))),
        }
    )

    df3 = pd.DataFrame(
        {
            "P": [1, 2, 3, 1, 2, 3],
            "L": [1, 1, 1, 1, 1, 1],
            "TARGET": [88, 1, 5, 99, 12, 11],
        }
    )

    df_list = [df1]
    new_partitions = [
        df2,
        df3,
    ]

    store_dataframes_as_dataset(
        dfs=df_list,
        store=store_factory,
        dataset_uuid="dataset_uuid",
        metadata_version=metadata_version,
    )

    bound_update_dataset(
        new_partitions,
        store=store_factory,
        dataset_uuid="dataset_uuid",
        metadata={"extra": "metadata"},
        default_metadata_version=metadata_version,
        sort_partitions_by=["TARGET"],
    )

    # Check that the `sort_partitions_by` column is indeed sorted monotonically among partitions
    for df in read_dataset_as_dataframes__iterator(
        store=store_factory, dataset_uuid="dataset_uuid"
    ):
        assert (df.TARGET == sorted(df.TARGET)).all()


@pytest.mark.xfail(
    reason="Unclear what this is actually testing, other than the mock itself"
)
def test_metadata_version(
    store_factory,
    bound_update_dataset,
    mock_default_metadata_version,
    backend_identifier,
):
    if backend_identifier in ("dask.dataframe", "dask.delayed"):
        pytest.skip()  # TODO: fix `io.dask.*.test_update._update_dataset`

    dataset_uuid = "dataset_uuid"
    partitions = [
        pd.DataFrame({"p": [1, 2]}),
        pd.DataFrame({"p": [2, 3]}),
    ]

    dataset = store_dataframes_as_dataset(
        dfs=partitions,
        store=store_factory,
        dataset_uuid=dataset_uuid,
        metadata_version=DEFAULT_METADATA_VERSION,
    )

    # with pytest.raises(AssertionError, match="Traversed through mock"):
    #     # Try to commit data to dataset using a different metadata version
    #     # and different data format (format is mocked)
    #     # This does not raise when the `parse_input_to_metapartition`
    #     # argument is `default_metadata_version` instead of `metadata_version`
    new_partitions = [pd.DataFrame({"p": [2, 3]})]
    bound_update_dataset(
        new_partitions,
        store=store_factory,
        dataset_uuid=dataset_uuid,
        default_metadata_version=mock_default_metadata_version,
    )

    mps = read_dataset_as_metapartitions(store=store_factory, dataset_uuid=dataset_uuid)
    assert len(mps) == len(dataset.partitions)


def test_raises_on_invalid_input(store_factory, bound_update_dataset):
    dataset_uuid = "dataset_uuid"
    partitions = [
        pd.DataFrame({"p": [1, 2]}),
        pd.DataFrame({"p": [2, 3]}),
    ]

    dataset = store_dataframes_as_dataset(
        dfs=partitions, store=store_factory, dataset_uuid=dataset_uuid
    )

    with pytest.raises(ValueError):
        new_partitions = [({"stuff"}, [("something", {1, 2, 3})])]  # invalid format
        bound_update_dataset(
            new_partitions, store=store_factory, dataset_uuid=dataset_uuid
        )

    # Check no new partitions have been written to storage
    mps = read_dataset_as_metapartitions(store=store_factory, dataset_uuid=dataset_uuid)
    assert len(mps) == len(dataset.partitions)


def test_update_secondary_indices_subset(store_factory, bound_update_dataset):
    df1 = pd.DataFrame({"A": range(10), "indexed": 1})
    dataset_uuid = "dataset_uuid"
    bound_update_dataset(
        df1, dataset_uuid=dataset_uuid, store=store_factory, secondary_indices="indexed"
    )

    df2 = pd.DataFrame({"A": range(10), "indexed": 2})
    # secondary index is omitted. plateau should pick it up regardless
    bound_update_dataset(df2, dataset_uuid=dataset_uuid, store=store_factory)

    dm = DatasetMetadata.load_from_store(
        dataset_uuid, store_factory(), load_all_indices=True
    )
    obs_values = dm.indices["indexed"].observed_values()

    assert sorted(obs_values) == [1, 2]

    with pytest.raises(ValueError, match="Incorrect indices provided"):
        # secondary index is omitted. plateau should pick it up regardless
        bound_update_dataset(
            df2, dataset_uuid=dataset_uuid, store=store_factory, secondary_indices="A"
        )


def test_update_first_time_with_secondary_indices(store_factory, bound_update_dataset):
    dataset_uuid = "dataset_uuid"
    index_column = "p"
    new_partition = [pd.DataFrame({index_column: [1, 2]})]

    bound_update_dataset(
        [new_partition],
        store=store_factory,
        dataset_uuid=dataset_uuid,
        secondary_indices=[index_column],
    )


def test_partition_on_null(store_factory, bound_update_dataset):  # gh-262
    keys = ["a", "b", "c", np.nan]
    values = range(len(keys))
    d = dict(zip(keys, values, strict=False))
    df = (
        pd.DataFrame.from_dict(d, orient="index")
        .reset_index()
        .rename(columns={"index": "part", 0: "value"})
    )

    with pytest.raises(
        Exception, match=r"Original dataframe size .* on a column with null values."
    ):
        bound_update_dataset(
            [df],
            store=store_factory,
            dataset_uuid="a_unique_dataset_identifier",
            partition_on=["part"],
        )


def test_update_infers_partition_on(store_factory, bound_update_dataset, df_not_nested):
    dataset_uuid = "dataset_uuid"

    dataset = bound_update_dataset(
        [df_not_nested],
        dataset_uuid=dataset_uuid,
        store=store_factory,
        partition_on=df_not_nested.columns[0],
    )
    # update the dataset
    # do not use partition_on since it should be interfered from the existing dataset

    updated_dataset = bound_update_dataset(
        [df_not_nested],
        dataset_uuid=dataset_uuid,
        store=store_factory,
    )

    assert len(updated_dataset.partitions) == 2 * len(dataset.partitions)


def test_update_raises_incompatible_partition_keys(
    store_factory, bound_update_dataset, df_not_nested
):
    dataset_uuid = "dataset_uuid"
    bound_update_dataset(
        [df_not_nested],
        dataset_uuid=dataset_uuid,
        store=store_factory,
        partition_on=df_not_nested.columns[0],
    )
    # Not allowed to use different partition_on
    with pytest.raises(
        ValueError, match="Incompatible set of partition keys encountered."
    ):
        bound_update_dataset(
            [df_not_nested],
            dataset_uuid=dataset_uuid,
            store=store_factory,
            partition_on=df_not_nested.columns[1],
        )


def test_update_raises_incompatible_inidces(
    store_factory, bound_update_dataset, df_not_nested
):
    dataset_uuid = "dataset_uuid"
    bound_update_dataset(
        [df_not_nested],
        dataset_uuid=dataset_uuid,
        store=store_factory,
        secondary_indices=df_not_nested.columns[0],
    )
    # Not allowed to update with indices which do not yet exist in dataset
    with pytest.raises(ValueError, match="indices"):
        bound_update_dataset(
            [df_not_nested],
            dataset_uuid=dataset_uuid,
            store=store_factory,
            secondary_indices=df_not_nested.columns[1],
        )


def test_update_of_dataset_with_non_default_table_name(
    store_factory, bound_update_dataset
):
    """Tests that datasets with table names other than "table" can be created,
    updated and read successfully (regression test for issue #445)."""

    # Create initial dataset
    dataset_uuid = "dataset_uuid"
    df_create = pd.DataFrame(
        {"date": [date(2021, 1, 1), date(2021, 1, 2)], "value": range(2)}
    )
    store_dataframes_as_dataset(
        dfs=[df_create],
        store=store_factory,
        dataset_uuid=dataset_uuid,
        table_name="non-default-name",
        partition_on=["date"],
    )
    dm = DatasetMetadata.load_from_store(dataset_uuid, store_factory())
    assert dm.table_name == "non-default-name"

    # Update dataset
    df_update = pd.DataFrame(
        {"date": [date(2021, 1, 3), date(2021, 1, 4)], "value": range(2)}
    )
    bound_update_dataset(
        [df_update],
        store=store_factory,
        dataset_uuid=dataset_uuid,
        table_name="non-default-name",
        partition_on=["date"],
    )
    dm = DatasetMetadata.load_from_store(dataset_uuid, store_factory())
    assert dm.table_name == "non-default-name"

    # Assert equality of dataframe
    df_read = (
        read_dataset_as_ddf(dataset_uuid, store_factory(), "table")
        .compute()
        .reset_index(drop=True)
    )
    df_expected = pd.concat([df_create, df_update]).reset_index(drop=True)
    pd.testing.assert_frame_equal(df_read, df_expected)


def _dtype_from_storage_nan_value(storage_backend, na_value):
    if PANDAS_3:
        dtype = pd.StringDtype(storage=storage_backend, na_value=na_value)
    else:
        if storage_backend == "pyarrow" and na_value is pd.NA:
            dtype = "string[pyarrow]"
        elif storage_backend == "pyarrow" and na_value is np.nan:
            dtype = "string[pyarrow_numpy]"
        elif storage_backend == "python" and na_value is np.nan:
            return None
        elif storage_backend == "python" and na_value is pd.NA:
            dtype = "string"
        else:
            raise ValueError(f"Unsupported storage backend: {storage_backend}")
    return dtype


@pytest.mark.parametrize("storage_backend", ["pyarrow", "python"])
@pytest.mark.parametrize("na_value", [np.nan, pd.NA])
def test_update_after_empty_partition_string_dtypes(
    store_factory, bound_update_dataset, storage_backend, na_value, backend_identifier
):
    import pandas as pd

    with pd.option_context("future.infer_string", True):
        other_nan_value = {np.nan, pd.NA}
        other_nan_value.remove(na_value)
        other_nan_value = other_nan_value.pop()
        dtype = _dtype_from_storage_nan_value(storage_backend, na_value)
        if dtype is None:
            pytest.skip()
        df = pd.DataFrame({"str": pd.Series(["a", "b", None], dtype=dtype)})

        dataset_uuid = "dataset_uuid"
        bound_update_dataset(
            [df.iloc[0:0]],  # empty partition
            store=store_factory,
            dataset_uuid=dataset_uuid,
        )
        # Schema verification should not fail
        bound_update_dataset(
            [df],
            store=store_factory,
            dataset_uuid=dataset_uuid,
        )
        if na_value is pd.NA:
            expected_dtype = _dtype_from_storage_nan_value("python", pd.NA)
        else:
            expected_dtype = _dtype_from_storage_nan_value("pyarrow", np.nan)
        # We have to cast to the expected dtype since pyarrow is only reading
        # the above two data types in. They are ignoring the written storage
        # backend and are defaulting to python for pd.NA and to pyarrow for
        # np.nan
        df["str"] = df["str"].astype(expected_dtype)

        pdt.assert_frame_equal(read_table(dataset_uuid, store_factory()), df)
        if backend_identifier == "dask.dataframe":
            # FIXME: dask.dataframe triggers the schema validation error but somehow
            # the exception is not properly forwarded and the test always fails
            return
        for storage in ["pyarrow", "python"]:
            df = pd.DataFrame(
                {
                    "str": pd.Series(
                        ["c", "d"],
                        dtype=_dtype_from_storage_nan_value(storage, other_nan_value),
                    )
                }
            )
            # Should be a ValueError but dask sometimes raises a different exception
            # type
            with pytest.raises(ValueError, match="Schemas.*are not compatible.*"):
                bound_update_dataset(
                    [df],
                    store=store_factory,
                    dataset_uuid=dataset_uuid,
                )


@pytest.mark.parametrize("storage_backend", ["pyarrow", "python"])
@pytest.mark.parametrize("na_value", [np.nan, pd.NA])
def test_update_after_empty_partition_string_dtypes_categoricals(
    store_factory, bound_update_dataset, storage_backend, na_value
):
    import pandas as pd

    with pd.option_context("future.infer_string", True):
        other_nan_value = {np.nan, pd.NA}
        other_nan_value.remove(na_value)
        other_nan_value = other_nan_value.pop()
        dtype = _dtype_from_storage_nan_value(storage_backend, na_value)
        if dtype is None:
            pytest.skip()
        df = pd.DataFrame(
            {"str": pd.Series(["a", "b", None], dtype=dtype).astype("category")}
        )

        dataset_uuid = "dataset_uuid"
        bound_update_dataset(
            [df.iloc[0:0]],  # empty partition
            store=store_factory,
            dataset_uuid=dataset_uuid,
        )
        # Schema verification should not fail
        bound_update_dataset(
            [df],
            store=store_factory,
            dataset_uuid=dataset_uuid,
        )
        expected_dtype = _dtype_from_storage_nan_value("pyarrow", np.nan)
        # We have to cast to the expected dtype since pyarrow is only reading
        # categoricals with the pyarrow_numpy data type.
        df["str"] = df["str"].astype(expected_dtype)

        pdt.assert_frame_equal(read_table(dataset_uuid, store_factory()), df)
        for storage in ["pyarrow", "python"]:
            df = pd.DataFrame(
                {
                    "str": pd.Series(
                        ["c", "d"],
                        dtype=_dtype_from_storage_nan_value(storage, other_nan_value),
                    ).astype("category")
                }
            )
            bound_update_dataset(
                [df],
                store=store_factory,
                dataset_uuid=dataset_uuid,
            )
    after_update = read_table(dataset_uuid, store_factory())

    if not PANDAS_3:
        expected_dtype = "object"

    expected_after_update = pd.DataFrame(
        {"str": pd.Series(["a", "b", None, "c", "d", "c", "d"], dtype=expected_dtype)}
    )
    pdt.assert_frame_equal(after_update, expected_after_update)

    # Storage of categorical dtypes will only happen with np.nan If we try the other na_value we'll get a validation error

    for storage in ["pyarrow", "python"]:
        df = pd.DataFrame(
            {
                "str": pd.Series(
                    ["e", "f", None],
                    dtype=_dtype_from_storage_nan_value(storage, pd.NA),
                )
            }
        )
        with pytest.raises(ValueError, match="Schemas.*are not compatible.*"):
            bound_update_dataset(
                [df],
                store=store_factory,
                dataset_uuid=dataset_uuid,
            )

    # With np.nan works fine?
    skipped = False
    for storage in ["pyarrow", "python"]:
        dtype = _dtype_from_storage_nan_value(storage, np.nan)
        if dtype is None:
            skipped = True
            continue
        df = pd.DataFrame(
            {
                "str": pd.Series(
                    ["e", "f", None],
                    dtype=dtype,
                )
            }
        )
        bound_update_dataset(
            [df],
            store=store_factory,
            dataset_uuid=dataset_uuid,
        )

    after_update_as_cats = read_table(
        dataset_uuid, store_factory(), categoricals=["str"]
    )
    values = ["a", "b", None, "c", "d", "c", "d", "e", "f", None, "e", "f", None]
    if skipped:
        values = values[:-3]
    expected = pd.DataFrame(
        {
            "str": pd.Series(
                values,
                dtype=expected_dtype,
            ).astype("category")
        }
    )
    pdt.assert_frame_equal(after_update_as_cats, expected)



import inspect
import os
import os.path
import sys

from sphinx.ext import apidoc

from plateau import __version__ as version

package = "plateau"

nitpicky = True

__location__ = os.path.join(
    os.getcwd(), os.path.dirname(inspect.getfile(inspect.currentframe()))
)

# Generate module references
output_dir = os.path.abspath(os.path.join(__location__, "../docs/_rst"))
module_dir = os.path.abspath(os.path.join(__location__, "..", package))

apidoc_parameters = ["-f", "-e", "-o", output_dir, module_dir]
apidoc.main(apidoc_parameters)

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), "sphinxext")))

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "reference_aliases",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.todo",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.ifconfig",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# General information about the project.
project = "plateau"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

release = version

html_static_path = ["_static"]

# Output file base name for HTML help builder.
htmlhelp_basename = "plateau-doc"


# -- External mapping ------------------------------------------------------------
python_version = ".".join(map(str, sys.version_info[0:2]))

nitpick_ignore = [
    ("py:class", "plateau.core._mixins.CopyMixin"),
    ("py:class", "plateau.serialization.LiteralValue"),
    ("py:class", "dask.delayed.Delayed"),
    # Literals
    ("py:class", "exact"),
    ("py:class", "all"),
    ("py:class", "prefix"),
    # typing.TypeVars not working
    ("py:class", "LiteralValue"),
    ("py:class", "T"),
    ("py:class", ".."),
    ("py:class", "ValueType"),
    # Caused by SchemaWrapper copying/overloading doc string of pyarrow
    ("py:class", "Field"),
    ("py:class", "Schema"),
    ("py:class", "default False"),
    ("py:class", "False"),
    ("py:class", "DataFrame"),
]

intersphinx_mapping = {
    "pandas": ("https://pandas.pydata.org/docs/", None),
    "minimalkv": ("https://minimalkv.readthedocs.io/en/stable/", None),
    "pyarrow": ("https://arrow.apache.org/docs/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "python": ("https://docs.python.org/3", None),
    "dask": ("https://docs.dask.org/en/stable/", None),
}

# In particular type annotations are rendered as its full path to the class but
# the sphinx docs usually expose the high level API path. This renames the
# sphinx targets appropriately. The replacements are all applied, therefore mind
# overlapping replacements.
reftarget_replace = {
    "pyarrow.lib.": "pyarrow.",
    "pd.": "pandas.",
    "pandas.core.series.Series": "pandas.Series",
    "pandas.core.frame.DataFrame": "pandas.DataFrame",
    "dask.dataframe.core.DataFrame": "dask.dataframe.DataFrame",
    "dask.dataframe.core.Series": "dask.dataframe.Series",
    "dask.bag.core.Bag": "dask.bag.Bag",
    "plateau.serialization._generic": "plateau.serialization",
    "plateau.serialization._parquet": "plateau.serialization",
}



import datetime
from collections import OrderedDict

import pandas as pd
import pytest
from pandas.testing import assert_frame_equal

from plateau.core.common_metadata import make_meta
from plateau.core.dataset import DatasetMetadata
from plateau.io.eager import (
    commit_dataset,
    create_empty_dataset_header,
    read_table,
    store_dataframes_as_dataset,
    write_single_partition,
)


def test_commit_dataset_from_metapartition(dataset_function, store):
    new_data = [
        pd.DataFrame(
            OrderedDict(
                [
                    ("P", [5]),
                    ("L", [5]),
                    ("TARGET", [5]),
                    ("DATE", [datetime.date(2016, 3, 23)]),
                ]
            )
        )
    ]
    new_partition = write_single_partition(
        store=store, dataset_uuid=dataset_function.uuid, data=new_data
    )
    pre_commit_dataset = DatasetMetadata.load_from_store(
        uuid=dataset_function.uuid, store=store
    )
    # Cannot assert equal since the metadata is differently ordered
    assert pre_commit_dataset == dataset_function

    updated_dataset = commit_dataset(
        store=store,
        dataset_uuid=dataset_function.uuid,
        new_partitions=new_partition,
        delete_scope=None,
        partition_on=None,
    )
    assert updated_dataset != dataset_function

    assert updated_dataset.uuid == dataset_function.uuid
    assert len(updated_dataset.partitions) == len(dataset_function.partitions) + 1

    # ensure that the new dataset is actually the one on disc
    loaded_dataset = DatasetMetadata.load_from_store(
        uuid=updated_dataset.uuid, store=store
    )
    assert loaded_dataset == updated_dataset

    # Read the data and check whether the rows above are included.
    # This checks whether all necessary information were updated in the header
    # (e.g. files attributes of the partitions)
    actual = read_table(store=store, dataset_uuid=dataset_function.uuid)
    df_expected = pd.DataFrame(
        OrderedDict(
            [
                (
                    "DATE",
                    [
                        datetime.date(2016, 3, 23),
                        datetime.date(2010, 1, 1),
                        datetime.date(2009, 12, 31),
                    ],
                ),
                ("L", [5, 1, 2]),
                ("P", [5, 1, 2]),
                ("TARGET", [5, 1, 2]),
            ]
        )
    )
    actual = actual.sort_values("DATE", ascending=False).reset_index(drop=True)

    assert_frame_equal(df_expected, actual)


# TODO: document changes for write_single_partition / commit workflow
def test_commit_dataset_from_nested_metapartition(store):
    """Check it is possible to use `commit_dataset` with nested metapartitions
    as input.

    Original issue: https://github.com/JDASoftwareGroup/plateau/issues/40
    """

    df = pd.DataFrame({"a": [1, 1, 2, 2], "b": [3, 4, 5, 6]})

    create_empty_dataset_header(
        store=store,
        dataset_uuid="uuid",
        schema=make_meta(df, "table", ["a"]),
        partition_on=["a"],
    )

    partitions = []
    for _ in range(2):
        partitions.append(
            write_single_partition(
                store=store,
                dataset_uuid="uuid",
                data=df,
                partition_on=["a"],
            )
        )

    partition_labels = {mp_.label for mp in partitions for mp_ in mp}
    dm = commit_dataset(
        store=store, dataset_uuid="uuid", new_partitions=partitions, partition_on=["a"]
    )
    assert dm.partitions.keys() == partition_labels


def test_initial_commit(store):
    dataset_uuid = "dataset_uuid"
    df = pd.DataFrame(OrderedDict([("P", [5]), ("L", [5]), ("TARGET", [5])]))
    dataset = create_empty_dataset_header(
        store=store,
        schema=make_meta(df, origin="1"),
        dataset_uuid=dataset_uuid,
        metadata_version=4,
    )
    assert dataset.explicit_partitions is False
    new_metapartition = write_single_partition(
        store=store, dataset_uuid=dataset.uuid, data=df
    )

    updated_dataset = commit_dataset(
        store=store,
        dataset_uuid=dataset.uuid,
        # FIXME: is this breaking and if so, is it expected?
        new_partitions=[new_metapartition],
        delete_scope=None,
        partition_on=None,
    )
    assert updated_dataset.explicit_partitions is True
    actual = read_table(store=store, dataset_uuid=updated_dataset.uuid)
    df_expected = pd.DataFrame(OrderedDict([("L", [5]), ("P", [5]), ("TARGET", [5])]))

    assert_frame_equal(df_expected, actual)


def test_commit_dataset_only_delete(store, metadata_version):
    partitions = [
        pd.DataFrame({"p": [1]}),
        pd.DataFrame({"p": [2]}),
    ]
    dataset = store_dataframes_as_dataset(
        dfs=partitions,
        store=lambda: store,
        metadata={"dataset": "metadata"},
        dataset_uuid="dataset_uuid",
        secondary_indices="p",
        metadata_version=metadata_version,
    )
    dataset = dataset.load_index("p", store)
    assert len(dataset.partitions) == 2

    delete_scope = [{"p": 1}]
    updated_dataset = commit_dataset(
        store=store,
        dataset_uuid=dataset.uuid,
        new_partitions=None,
        delete_scope=delete_scope,
        partition_on=None,
    )
    assert len(updated_dataset.partitions) == 1
    assert updated_dataset.explicit_partitions is True


def test_commit_dataset_delete_all(store, metadata_version):
    partitions = [pd.DataFrame({"p": [1]})]

    dataset = store_dataframes_as_dataset(
        dfs=partitions,
        store=lambda: store,
        metadata={"dataset": "metadata"},
        dataset_uuid="dataset_uuid",
        secondary_indices="p",
        metadata_version=metadata_version,
    )
    dataset = dataset.load_index("p", store)
    assert len(dataset.partitions) == 1

    delete_scope = [{"p": 1}]
    updated_dataset = commit_dataset(
        store=store,
        dataset_uuid=dataset.uuid,
        new_partitions=None,
        delete_scope=delete_scope,
        partition_on=None,
    )
    assert len(updated_dataset.partitions) == 0
    assert updated_dataset.explicit_partitions is True


def test_commit_dataset_add_metadata(store, metadata_version):
    df = pd.DataFrame({"p": [1]})
    dataset = store_dataframes_as_dataset(
        dfs=[df],
        store=lambda: store,
        metadata={"old": "metadata"},
        dataset_uuid="dataset_uuid",
        metadata_version=metadata_version,
    )

    commit_dataset(store=store, dataset_uuid=dataset.uuid, metadata={"new": "metadata"})
    restored_ds = DatasetMetadata.load_from_store(uuid="dataset_uuid", store=store)

    actual_metadata = restored_ds.metadata
    del actual_metadata["creation_time"]
    expected_meta = {"new": "metadata", "old": "metadata"}
    assert actual_metadata == expected_meta


def test_commit_dataset_raises_no_input(store):
    with pytest.raises(ValueError, match="Need to provide either"):
        commit_dataset(store=store, dataset_uuid="something")

    with pytest.raises(ValueError, match="Need to provide either"):
        commit_dataset(store=store, dataset_uuid="something", new_partitions=[])
    with pytest.raises(ValueError, match="Need to provide either"):
        commit_dataset(store=store, dataset_uuid="something", metadata={})
    with pytest.raises(ValueError, match="Need to provide either"):
        commit_dataset(store=store, dataset_uuid="something", delete_scope=[])



import dask
import dask.dataframe as dd
import pandas as pd
import pandas.testing as pdt
import pytest

from plateau.io.dask.compression import (
    pack_payload,
    pack_payload_pandas,
    unpack_payload,
    unpack_payload_pandas,
)


def test_pack_payload(df_all_types):
    # For a single row dataframe the packing actually has a few more bytes
    with dask.config.set({"dataframe.convert-string": False}):
        df = dd.from_pandas(
            pd.concat([df_all_types] * 10, ignore_index=True), npartitions=3
        )
    size_before = df.memory_usage(deep=True).sum()

    packed_df = pack_payload(df, group_key=list(df.columns[-2:]))

    size_after = packed_df.memory_usage(deep=True).sum()

    assert (size_after < size_before).compute()


def test_pack_payload_empty(df_all_types):
    # For a single row dataframe the packing actually has a few more bytes
    df_empty = dd.from_pandas(df_all_types.iloc[:0], npartitions=1)

    group_key = [df_all_types.columns[-1]]
    pdt.assert_frame_equal(
        df_empty.compute(),
        unpack_payload(
            pack_payload(df_empty, group_key=group_key), unpack_meta=df_empty._meta
        ).compute(),
    )


def test_pack_payload_pandas(df_all_types):
    # For a single row dataframe the packing actually has a few more bytes
    df = pd.concat([df_all_types] * 10, ignore_index=True)
    size_before = df.memory_usage(deep=True).sum()

    packed_df = pack_payload_pandas(df, group_key=list(df.columns[-2:]))

    size_after = packed_df.memory_usage(deep=True).sum()

    assert size_after < size_before


def test_pack_payload_pandas_empty(df_all_types):
    # For a single row dataframe the packing actually has a few more bytes
    df_empty = df_all_types.iloc[:0]

    group_key = [df_all_types.columns[-1]]
    pdt.assert_frame_equal(
        df_empty,
        unpack_payload_pandas(
            pack_payload_pandas(df_empty, group_key=group_key), unpack_meta=df_empty
        ),
    )


@pytest.mark.parametrize("num_group_cols", [1, 4])
def test_pack_payload_roundtrip(df_all_types, num_group_cols):
    group_key = list(df_all_types.columns[-num_group_cols:])
    with dask.config.set({"dataframe.convert-string": False}):
        df_all_types = dd.from_pandas(df_all_types, npartitions=2)
    pdt.assert_frame_equal(
        df_all_types.compute(),
        unpack_payload(
            pack_payload(df_all_types, group_key=group_key),
            unpack_meta=df_all_types._meta,
        ).compute(),
    )


@pytest.mark.parametrize("num_group_cols", [1, 4])
def test_pack_payload_pandas_roundtrip(df_all_types, num_group_cols):
    group_key = list(df_all_types.columns[-num_group_cols:])
    pdt.assert_frame_equal(
        df_all_types,
        unpack_payload_pandas(
            pack_payload_pandas(df_all_types, group_key=group_key),
            unpack_meta=df_all_types,
        ),
    )



"""This module contains logic to update an existing dataset.

Update means adding new partitions and deleting existing partitions.
plateau does not allow an update of the content of existing partitions.
"""

from plateau.core.index import PartitionIndex
from plateau.core.utils import ensure_store
from plateau.io_components.write import store_dataset_from_partitions


def _get_partitions(dataset, query_params):
    partitions = []
    for params in query_params:
        partitions += dataset.query(**params)

    return partitions


def update_dataset_from_partitions(
    partition_list,
    store_factory,
    dataset_uuid,
    ds_factory,
    delete_scope,
    metadata,
    metadata_merger,
):
    store = ensure_store(store_factory)

    if ds_factory:
        ds_factory = ds_factory.load_all_indices()
        remove_partitions = _get_partitions(ds_factory, delete_scope)

        index_columns = list(ds_factory.indices.keys())
        for column in index_columns:
            index = ds_factory.indices[column]
            if isinstance(index, PartitionIndex):
                del ds_factory.indices[column]
    else:
        # Dataset does not exist yet.
        remove_partitions = []

    new_dataset = store_dataset_from_partitions(
        partition_list=partition_list,
        store=store,
        dataset_uuid=dataset_uuid,
        dataset_metadata=metadata,
        metadata_merger=metadata_merger,
        update_dataset=ds_factory,
        remove_partitions=remove_partitions,
    )

    return new_dataset



from collections.abc import Sequence
from functools import partial

import dask.bag as db
from dask.delayed import Delayed

from plateau.core import naming
from plateau.core.docs import default_docs
from plateau.core.factory import DatasetFactory, _ensure_factory
from plateau.core.typing import StoreInput
from plateau.core.utils import lazy_store
from plateau.core.uuid import gen_uuid
from plateau.io.dask._utils import (
    _cast_categorical_to_index_cat,
    _get_data,
    _maybe_get_categoricals_from_index,
)
from plateau.io_components.index import update_indices_from_partitions
from plateau.io_components.metapartition import (
    SINGLE_TABLE,
    MetaPartition,
    parse_input_to_metapartition,
)
from plateau.io_components.read import dispatch_metapartitions_from_factory
from plateau.io_components.utils import normalize_args, raise_if_indices_overlap
from plateau.io_components.write import (
    raise_if_dataset_exists,
    store_dataset_from_partitions,
)

__all__ = (
    "read_dataset_as_dataframe_bag",
    "store_bag_as_dataset",
    "build_dataset_indices__bag",
)


def _store_dataset_from_partitions_flat(mpss, *args, **kwargs):
    return store_dataset_from_partitions(
        [mp for sublist in mpss for mp in sublist], *args, **kwargs
    )


def _load_and_concat_metapartitions_inner(mps, *args, **kwargs):
    return MetaPartition.concat_metapartitions(
        [mp.load_dataframes(*args, **kwargs) for mp in mps]
    )


@default_docs
def read_dataset_as_metapartitions_bag(
    dataset_uuid=None,
    store=None,
    columns=None,
    predicate_pushdown_to_io=True,
    categoricals=None,
    dates_as_object: bool = True,
    predicates=None,
    factory=None,
    dispatch_by=None,
    partition_size=None,
):
    """Retrieve dataset as `dask.bag.Bag` of `MetaPartition` objects.

    Parameters
    ----------

    Returns
    -------
    dask.bag.Bag:
        A dask.bag object containing the metapartions.
    """
    ds_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
    )

    store = ds_factory.store_factory
    mps = dispatch_metapartitions_from_factory(
        dataset_factory=ds_factory,
        predicates=predicates,
        dispatch_by=dispatch_by,
    )
    mp_bag = db.from_sequence(mps, partition_size=partition_size)

    if dispatch_by is not None:
        mp_bag = mp_bag.map(
            _load_and_concat_metapartitions_inner,
            store=store,
            columns=columns,
            categoricals=categoricals,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            dates_as_object=dates_as_object,
            predicates=predicates,
        )
    else:
        mp_bag = mp_bag.map(
            MetaPartition.load_dataframes,
            store=store,
            columns=columns,
            categoricals=categoricals,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            dates_as_object=dates_as_object,
            predicates=predicates,
        )

    categoricals_from_index = _maybe_get_categoricals_from_index(
        ds_factory, categoricals
    )

    if categoricals_from_index:
        mp_bag = mp_bag.map(
            MetaPartition.apply,
            func=partial(
                _cast_categorical_to_index_cat, categories=categoricals_from_index
            ),
            type_safe=True,
        )
    return mp_bag


@default_docs
def read_dataset_as_dataframe_bag(
    dataset_uuid=None,
    store=None,
    columns=None,
    predicate_pushdown_to_io=True,
    categoricals=None,
    dates_as_object: bool = True,
    predicates=None,
    factory=None,
    dispatch_by=None,
    partition_size=None,
):
    """Retrieve data as dataframe from a :class:`dask.bag.Bag` of
    `MetaPartition` objects.

    Parameters
    ----------

    Returns
    -------
    dask.bag.Bag
        A dask.bag.Bag which contains the metapartitions and mapped to a function for retrieving the data.
    """
    mps = read_dataset_as_metapartitions_bag(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
        columns=columns,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        categoricals=categoricals,
        dates_as_object=dates_as_object,
        predicates=predicates,
        dispatch_by=dispatch_by,
        partition_size=partition_size,
    )
    return mps.map(_get_data)


@default_docs
@normalize_args
def store_bag_as_dataset(
    bag,
    store,
    dataset_uuid=None,
    metadata=None,
    df_serializer=None,
    overwrite=False,
    metadata_merger=None,
    metadata_version=naming.DEFAULT_METADATA_VERSION,
    partition_on=None,
    metadata_storage_format=naming.DEFAULT_METADATA_STORAGE_FORMAT,
    secondary_indices=None,
    table_name: str = SINGLE_TABLE,
):
    """Transform and store a dask.bag of dictionaries containing dataframes to
    a plateau dataset in store.

    This is the dask.bag-equivalent of
    :func:`~plateau.io.dask.delayed.store_delayed_as_dataset`. See there
    for more detailed documentation on the different possible input types.

    Parameters
    ----------
    bag: dask.bag.Bag
        A dask bag containing dictionaries of dataframes or dataframes.
    """
    store = lazy_store(store)
    if dataset_uuid is None:
        dataset_uuid = gen_uuid()

    if not overwrite:
        raise_if_dataset_exists(dataset_uuid=dataset_uuid, store=store)

    raise_if_indices_overlap(partition_on, secondary_indices)

    input_to_mps = partial(
        parse_input_to_metapartition,
        metadata_version=metadata_version,
        table_name=table_name,
    )
    mps = bag.map(input_to_mps)

    if partition_on:
        mps = mps.map(MetaPartition.partition_on, partition_on=partition_on)

    if secondary_indices:
        mps = mps.map(MetaPartition.build_indices, columns=secondary_indices)

    mps = mps.map(
        MetaPartition.store_dataframes,
        store=store,
        df_serializer=df_serializer,
        dataset_uuid=dataset_uuid,
    )

    aggregate = partial(
        _store_dataset_from_partitions_flat,
        dataset_uuid=dataset_uuid,
        store=store,
        dataset_metadata=metadata,
        metadata_merger=metadata_merger,
        metadata_storage_format=metadata_storage_format,
    )

    return mps.reduction(perpartition=list, aggregate=aggregate, split_every=False)


@default_docs
def build_dataset_indices__bag(
    store: StoreInput | None,
    dataset_uuid: str | None,
    columns: Sequence[str],
    partition_size: int | None = None,
    factory: DatasetFactory | None = None,
) -> Delayed:
    """Function which builds a
    :class:`~plateau.core.index.ExplicitSecondaryIndex`.

    This function loads the dataset, computes the requested indices and writes
    the indices to the dataset. The dataset partitions itself are not mutated.

    Parameters
    ----------
    """
    ds_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
    )

    assert ds_factory.schema is not None
    cols_to_load = set(columns) & set(ds_factory.schema.names)

    mps = dispatch_metapartitions_from_factory(ds_factory)

    return (
        db.from_sequence(seq=mps, partition_size=partition_size)
        .map(
            MetaPartition.load_dataframes,
            store=ds_factory.store_factory,
            columns=cols_to_load,
        )
        .map(MetaPartition.build_indices, columns=columns)
        .map(MetaPartition.remove_dataframes)
        .reduction(list, list, split_every=False, out_type=db.Bag)
        .flatten()
        .map_partitions(list)
        .map_partitions(
            update_indices_from_partitions, dataset_metadata_factory=ds_factory
        )
    )



import pandas as pd
import pyarrow as pa
import simplejson
from packaging.version import parse as parse_version
from pandas.errors import OptionError

PANDAS_3 = parse_version(pd.__version__).major >= 3

ARROW_GE_20 = parse_version(pa.__version__).major >= 20


def pandas_infer_string() -> bool:
    if parse_version(pd.__version__).major >= 3:
        # In pandas 3, infer_string is always True
        return True
    try:
        return pd.get_option("future.infer_string")
    except OptionError:
        return False


def load_json(buf, **kwargs):
    """Compatibility function to load JSON from str/bytes/unicode.

    For Python 2.7 json.loads accepts str and unicode. Python 3.4 only
    accepts str whereas 3.5+ accept bytes and str.
    """
    if isinstance(buf, bytes):
        return simplejson.loads(buf.decode("utf-8"), **kwargs)
    else:
        return simplejson.loads(buf, **kwargs)



import pickle
from functools import partial

import dask
import dask.dataframe as dd
import numpy as np
import pandas as pd
import pytest
from dask.dataframe.utils import assert_eq as assert_dask_eq
from packaging import version
from pandas import testing as pdt
from pandas.testing import assert_frame_equal

from plateau.core.testing import get_dataframe_not_nested
from plateau.io.dask.dataframe import read_dataset_as_ddf
from plateau.io.eager import store_dataframes_as_dataset
from plateau.io.testing.read import *  # noqa
from plateau.io_components.metapartition import SINGLE_TABLE

PANDAS_LT_3 = version.parse(pd.__version__) < version.parse("3.0.0.dev")


@pytest.fixture()
def output_type():
    return "table"


def _read_as_ddf(
    dataset_uuid,
    store,
    factory=None,
    categoricals=None,
    tables=None,
    dataset_has_index=False,
    **kwargs,
):
    table = tables or SINGLE_TABLE

    ddf = read_dataset_as_ddf(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
        categoricals=categoricals,
        table=table,
        **kwargs,
    )
    if categoricals:
        assert ddf._meta.dtypes["P"] == pd.api.types.CategoricalDtype(
            categories=["__UNKNOWN_CATEGORIES__"], ordered=False
        )
        if dataset_has_index:
            assert ddf._meta.dtypes["L"] == pd.api.types.CategoricalDtype(
                categories=[1, 2], ordered=False
            )
        else:
            assert ddf._meta.dtypes["L"] == pd.api.types.CategoricalDtype(
                categories=["__UNKNOWN_CATEGORIES__"], ordered=False
            )

    s = pickle.dumps(ddf, pickle.HIGHEST_PROTOCOL)
    ddf = pickle.loads(s)

    ddf = ddf.compute().reset_index(drop=True)

    def extract_dataframe(ix):
        df = ddf.iloc[[ix]].copy()
        for col in df.columns:
            if isinstance(df[col].dtype, pd.CategoricalDtype):
                df[col] = df[col].cat.remove_unused_categories()
        return df.reset_index(drop=True)

    return [extract_dataframe(ix) for ix in ddf.index]


@pytest.fixture()
def bound_load_dataframes():
    return _read_as_ddf


def test_load_dataframe_categoricals_with_index(dataset_with_index_factory):
    func = partial(_read_as_ddf, dataset_has_index=True)
    test_read_dataset_as_dataframes(  # noqa: F405
        dataset_factory=dataset_with_index_factory,
        dataset=dataset_with_index_factory,
        store_session_factory=dataset_with_index_factory.store_factory,
        use_dataset_factory=True,
        bound_load_dataframes=func,
        use_categoricals=True,
        output_type="table",
        dates_as_object=False,
    )


def test_read_ddf_from_categorical_partition(store_factory):
    df = pd.DataFrame({"x": ["a"]}).astype({"x": "category"})
    store_dataframes_as_dataset(
        dfs=[df], dataset_uuid="dataset_uuid", store=store_factory
    )
    ddf = read_dataset_as_ddf(
        dataset_uuid="dataset_uuid", store=store_factory, table="table"
    )
    df_expected = pd.DataFrame({"x": ["a"]})
    df_actual = ddf.compute(scheduler="sync")
    pdt.assert_frame_equal(df_expected, df_actual)

    ddf = read_dataset_as_ddf(
        dataset_uuid="dataset_uuid",
        store=store_factory,
        categoricals=["x"],
        table="table",
    )
    df_actual = ddf.compute(scheduler="sync")
    pdt.assert_frame_equal(df, df_actual)


@pytest.mark.parametrize("index_type", ["primary", "secondary"])
def test_reconstruct_dask_index(store_factory, index_type, monkeypatch):
    dataset_uuid = "dataset_uuid"
    colA = "ColumnA"
    colB = "ColumnB"
    df1 = pd.DataFrame({colA: [1, 2], colB: ["x", "y"]})
    df2 = pd.DataFrame({colA: [3, 4], colB: ["x", "y"]})
    df_chunks = np.array_split(pd.concat([df1, df2]).reset_index(drop=True), 4)
    if not PANDAS_LT_3:
        # Workaround for https://github.com/numpy/numpy/issues/24889.
        df_chunks = [
            pd.DataFrame(data=arr, columns=df1.columns).astype(df1.dtypes)
            for arr in df_chunks
        ]
    with dask.config.set(
        {"dataframe.convert-string": False, "dataframe.shuffle.method": "tasks"}
    ):
        ddf_expected = dd.from_map(lambda x: x, df_chunks).set_index(
            colA, divisions=[1, 2, 3, 4, 4]
        )
        ddf_expected_simple = dd.from_pandas(
            pd.concat([df1, df2]), npartitions=2
        ).set_index(colA)

    if index_type == "secondary":
        secondary_indices = colA
        partition_on = None
    else:
        secondary_indices = None
        partition_on = colA

    store_dataframes_as_dataset(
        store=store_factory,
        dataset_uuid=dataset_uuid,
        dfs=[df1, df2],
        secondary_indices=secondary_indices,
        partition_on=partition_on,
    )

    ddf = read_dataset_as_ddf(
        dataset_uuid=dataset_uuid,
        store=store_factory,
        table="table",
        dask_index_on=colA,
    )

    assert ddf_expected.npartitions == 4
    assert len(ddf_expected.divisions) == 5
    assert ddf_expected.divisions == (1, 2, 3, 4, 4)
    assert ddf.index.name == colA

    assert ddf.npartitions == 4
    assert len(ddf.divisions) == 5
    assert ddf.divisions == (1, 2, 3, 4, 4)

    assert_dask_eq(ddf_expected, ddf, scheduler="distributed")
    assert_dask_eq(
        ddf_expected_simple, ddf, check_divisions=False, scheduler="distributed"
    )

    assert_frame_equal(ddf_expected.compute(), ddf.compute())
    assert_frame_equal(ddf_expected_simple.compute(), ddf.compute())


@pytest.fixture()
def setup_reconstruct_dask_index_types(store_factory, df_not_nested):
    indices = list(df_not_nested.columns)
    indices.remove("null")
    return store_dataframes_as_dataset(
        store=store_factory,
        dataset_uuid="dataset_uuid",
        dfs=[df_not_nested],
        secondary_indices=indices,
    )


@pytest.mark.parametrize("col", get_dataframe_not_nested().columns)
def test_reconstruct_dask_index_types(
    store_factory, setup_reconstruct_dask_index_types, col
):
    if col == "null":
        pytest.xfail(reason="Cannot index null column")
    ddf = read_dataset_as_ddf(
        dataset_uuid=setup_reconstruct_dask_index_types.uuid,
        store=store_factory,
        table="table",
        dask_index_on=col,
    )
    assert ddf.known_divisions
    assert ddf.index.name == col


def test_reconstruct_dask_index_sorting(store_factory, monkeypatch):
    dataset_uuid = "dataset_uuid"
    colA = "ColumnA"
    colB = "ColumnB"

    df = pd.DataFrame(
        {colA: np.random.randint(high=100000, low=-100000, size=(50,)), colB: 0}
    )
    store_dataframes_as_dataset(
        store=store_factory, dataset_uuid=dataset_uuid, dfs=[df], partition_on=colA
    )
    ddf = read_dataset_as_ddf(
        dataset_uuid=dataset_uuid,
        store=store_factory,
        table="table",
        dask_index_on=colA,
    )

    assert all(
        ddf.map_partitions(lambda df: df.index.min()).compute().values
        == ddf.divisions[:-1]
    )


def test_reconstruct_dask_index_raise_no_index(store_factory):
    dataset_uuid = "dataset_uuid"
    colA = "ColumnA"
    df1 = pd.DataFrame({colA: [1, 2]})
    store_dataframes_as_dataset(
        store=store_factory, dataset_uuid=dataset_uuid, dfs=[df1]
    )
    with pytest.raises(
        RuntimeError,
        match=r"Requested index: \['ColumnA'\] but available index columns: \[\]",
    ):
        read_dataset_as_ddf(
            dataset_uuid=dataset_uuid,
            store=store_factory,
            table="table",
            dask_index_on=colA,
        )


def test_column_projection(store_factory, monkeypatch):
    dataset_uuid = "dataset_uuid"
    df1 = pd.DataFrame({"colA": [1, 2], "colB": ["x", "y"]})
    fake_called = False

    class FakeParquet:
        @classmethod
        def restore_dataframe(cls, store, key, filter_query, columns, *args, **kwargs):
            nonlocal fake_called
            fake_called = True
            assert columns == ["colA"]
            return df1[columns]

    from plateau.serialization import DataFrameSerializer

    monkeypatch.setitem(
        DataFrameSerializer._serializers,
        ".parquet",
        FakeParquet,
    )
    store_dataframes_as_dataset(
        store=store_factory, dataset_uuid=dataset_uuid, dfs=[df1]
    )
    ddf_auto = read_dataset_as_ddf(
        dataset_uuid=dataset_uuid,
        store=store_factory,
    )["colA"]
    ddf_manual = read_dataset_as_ddf(
        dataset_uuid=dataset_uuid, store=store_factory, columns=["colA"]
    )["colA"]
    assert_dask_eq(ddf_auto, ddf_manual)
    assert fake_called


def test_dask_index_on_non_string_raises(store_factory):
    dataset_uuid = "dataset_uuid"
    colA = 1
    df1 = pd.DataFrame({colA: [1, 2]})
    store_dataframes_as_dataset(
        store=store_factory, dataset_uuid=dataset_uuid, dfs=[df1]
    )
    with pytest.raises(
        TypeError,
        match=f"The parameter `dask_index_on` must be a string but got {type(colA)}",
    ):
        read_dataset_as_ddf(
            dataset_uuid=dataset_uuid,
            store=store_factory,
            table="table",
            dask_index_on=colA,
        )


def test_dask_dispatch_by_raises_if_index_on_not_none(store_factory):
    dataset_uuid = "dataset_uuid"
    colA = "ColumnA"
    df1 = pd.DataFrame({colA: [1, 2]})
    store_dataframes_as_dataset(
        store=store_factory, dataset_uuid=dataset_uuid, dfs=[df1]
    )
    with pytest.raises(
        ValueError,
        match="`read_dataset_as_ddf` got parameters `dask_index_on` and `dispatch_by`. "
        "Note that `dispatch_by` can only be used if `dask_index_on` is None.",
    ):
        read_dataset_as_ddf(
            dataset_uuid=dataset_uuid,
            store=store_factory,
            table="table",
            dask_index_on=colA,
            dispatch_by=[colA],
        )



import pytest

from plateau.serialization._util import ensure_unicode_string_type


@pytest.mark.parametrize("obj,expected", [("tüst", "tüst"), ("tüst".encode(), "tüst")])
def test_ensure_unicode_string_types(obj, expected):
    actual = ensure_unicode_string_type(obj)
    assert isinstance(actual, str)
    assert actual == expected



from plateau.core import *  # noqa: F401, F403
from plateau.core.utils import *  # noqa: F401, F403
from plateau.io.dask.bag import *  # noqa: F401, F403
from plateau.io.dask.dataframe import *  # noqa: F401, F403
from plateau.io.dask.delayed import *  # noqa: F401, F403
from plateau.io.eager import *  # noqa: F401, F403
from plateau.io.iter import *  # noqa: F401, F403



import os
import uuid

import pandas.testing as pdt
import pytest
from minimalkv import get_store_from_url

from plateau.core.testing import get_dataframe_alltypes
from plateau.serialization import ParquetSerializer

KNOWN_ARROW_VERSIONS = [
    "4.0.1",
    "5.0.0",
    "6.0.1",
    "7.0.0",
    "8.0.1",
    "9.0.0",
    "10.0.1",
    "11.0.0",
    "12.0.0",
    "13.0.0",
    "14.0.1",
    "14.0.2",
    "15.0.0",
    "15.0.2",
    "16.1.0",
    "17.0.0",
    "18.1.0",
    "19.0.0",
    "19.0.1",
    "20.0.0",
    "21.0.0",
]


@pytest.fixture(params=KNOWN_ARROW_VERSIONS)
def arrow_version(request):
    yield request.param


@pytest.fixture
def reference_store():
    path = os.path.join(
        os.path.dirname(os.path.dirname(os.path.realpath(__file__))),
        "..",
        "reference-data",
        "arrow-compat",
    )
    return get_store_from_url(f"hfs://{path}")


def test_current_arrow_version_tested():
    """Ensure that we do not forget to generate the reference file."""
    import pyarrow as pa
    from packaging.version import parse

    version = parse(pa.__version__)
    is_stable = not version.is_devrelease and not version.is_prerelease
    if is_stable:
        assert pa.__version__ in KNOWN_ARROW_VERSIONS


def test_arrow_compat(arrow_version, reference_store, mocker):
    """Test if reading/writing across the supported arrow versions is actually
    compatible.

    Generate new reference files by going to the `reference-data/arrow-
    compat` directory and executing `generate_reference.py` or
    `batch_generate_reference.sh`.
    """

    uuid_hook = mocker.patch("plateau.core.uuid._uuid_hook_object")
    uuid_hook.return_value = uuid.UUID(
        bytes=b"\x82\xd6\xc1\x06Z\x08\x11\xe9\x85eJ\x00\x07\xf8\n\x10"
    )

    orig = get_dataframe_alltypes()
    restored = ParquetSerializer().restore_dataframe(
        store=reference_store, key=arrow_version + ".parquet", date_as_object=True
    )

    pdt.assert_frame_equal(orig, restored)



import pickle
from copy import copy
from functools import partial

import pytest
from minimalkv import KeyValueStore

from plateau.core.factory import DatasetFactory


class CountStore(KeyValueStore):
    def __init__(self, inner):
        self.inner = inner
        self.get_count = 0

    def get(self, key):
        self.get_count += 1
        return self.inner.get(key)

    def iter_keys(self, prefix):
        return iter(self)

    def __iter__(self):
        return self.inner.__iter__()


class CountFactory:
    def __init__(self, inner):
        self.inner = inner
        self.count = 0
        self.last = None

    def __call__(self):
        self.count += 1
        result = self.inner()
        self.last = result
        return result

    def __getstate__(self):
        state = copy(self.__dict__)
        state["last"] = None
        return state


def _create_count_store(store_factory):
    return CountStore(store_factory())


@pytest.fixture(scope="function")
def count_store(store_factory):
    return CountFactory(partial(_create_count_store, store_factory))


def test_store_init(count_store, dataset_function):
    factory = DatasetFactory(dataset_uuid="dataset_uuid", store_factory=count_store)
    assert count_store.count == 0

    store = factory.store
    assert hasattr(store, "get")

    assert count_store.count == 1
    assert count_store.last == store

    assert store.get_count == 0

    # second get should cache
    store = factory.store
    assert count_store.count == 1
    assert count_store.last == store


def test_uuid(count_store, dataset_function):
    factory = DatasetFactory(dataset_uuid="dataset_uuid", store_factory=count_store)
    assert factory.dataset_uuid == "dataset_uuid"


def test_get_metadata(count_store, dataset_function):
    factory = DatasetFactory(dataset_uuid="dataset_uuid", store_factory=count_store)
    store = factory.store
    assert store.get_count == 0

    metadata = factory.dataset_metadata
    assert hasattr(metadata, "metadata")

    initial_count = store.get_count

    # second get should cache
    metadata = factory.dataset_metadata
    assert store.get_count == initial_count


def test_pickle(count_store, dataset_function):
    factory = DatasetFactory(dataset_uuid="dataset_uuid", store_factory=count_store)
    assert factory._cache_store is None
    assert factory._cache_metadata is None

    factory.store
    factory.dataset_metadata
    assert factory._cache_store is not None
    assert factory._cache_metadata is not None

    factory2 = pickle.loads(pickle.dumps(factory, pickle.HIGHEST_PROTOCOL))
    assert factory2._cache_store is None
    assert factory2._cache_metadata is None



import pytest

from plateau.io.eager import delete_dataset
from plateau.io.testing.delete import *  # noqa


def _delete_store_factory(dataset_uuid, store_factory):
    delete_dataset(dataset_uuid, store_factory)


def _delete_store(dataset_uuid, store_factory):
    delete_dataset(dataset_uuid, store_factory())


@pytest.fixture(params=["factory", "store-factory"])
def bound_delete_dataset(request):
    if request.param == "factory":
        return _delete_store_factory
    else:
        return _delete_store



"""Improved IO buffering compared to ``io.BufferedReader``.

The main issues w/ ``io.BufferedReader`` is that it is only meant of
sequential reads and that it resets the buffer on ``.seek(...)``. This
happens quite often in pyarrow and basically renders the buffereing
inefficient.
"""

import io
import logging
from typing import Any

_logger = logging.getLogger(__name__)


class BufferReadError(IOError):
    """Internal plateau error while attempting to read from buffer."""

    pass


class BlockBuffer(io.BufferedIOBase):
    """Block-based buffer.

    The input is split into fixed sizes blocks. Every block can be read
    independently.
    """

    def __init__(self, raw, blocksize=1024):
        self._raw = raw
        self._blocksize = blocksize
        self._size: int | None = None
        self._cached_blocks: list[Any] | None = None
        self._pos = 0

        if self._raw_closed():
            raise ValueError("Cannot use closed file object")
        if not self._raw_readable():
            raise ValueError("raw must be readable")
        if not self._raw_seekable():
            raise ValueError("raw must be seekable")
        if blocksize < 1:
            raise ValueError("blocksize must be at least 1")

    def _raw_closed(self):
        """If supported by ``raw``, return its closed state, otherwise return
        ``False``."""
        if hasattr(self._raw, "closed"):
            return self._raw.closed
        else:
            return False

    def _raw_readable(self):
        """If supported by ``raw``, return its readable state, otherwise return
        ``True``."""
        if hasattr(self._raw, "readable"):
            return self._raw.readable()
        else:
            return True

    def _raw_seekable(self):
        """If supported by ``raw``, return its seekable state, otherwise return
        ``True``."""
        if hasattr(self._raw, "seekable"):
            return self._raw.seekable()
        else:
            return True

    def _setup_cache(self):
        """Set up cache data structure and inspect underlying IO object.

        If the cache is already initialized, this is a no-op.
        """
        if self._cached_blocks is not None:
            # cache initialized, nothing to do
            return

        if hasattr(self._raw, "size"):
            self._size = self._raw.size
        elif hasattr(self._raw, "__len__"):
            self._size = len(self._raw)
        else:
            self._raw.seek(0, 2)
            self._size = self._raw.tell()

        n_blocks = self._size // self._blocksize
        if self._size % self._blocksize:
            n_blocks += 1

        self._cached_blocks = [None] * n_blocks

    def _fetch_blocks(self, block, n):
        """Fetch blocks from underlying IO object.

        This will mark the fetched blocks as loaded.

        Parameters
        ----------
        block: int
            First block to fetch.
        n: int
            Number of blocks to fetch.
        """
        assert n > 0

        # seek source
        offset = self._blocksize * block
        self._raw.seek(offset, 0)

        # read data into temporary variable and dump it into cache
        size = min(self._blocksize * n, self._size - offset)
        data = self._raw.read(size)
        if len(data) != size:
            err = (
                f"Expected raw read to return {size} bytes, but instead got {len(data)}"
            )
            _logger.error(err)
            raise BufferReadError(err)

        # fill blocks
        assert self._cached_blocks is not None
        for i in range(n):
            begin = i * self._blocksize
            end = min((i + 1) * self._blocksize, size)
            self._cached_blocks[block + i] = data[begin:end]

    def _ensure_range_loaded(self, start, size):
        """Ensure that a given byte range is loaded into the cache.

        This will scan for blocks that are not loaded yet and tries to load consecutive blocks as once.

        Parameters
        ----------
        start: int
            First byte of the range.
        size: int
            Number of bytes in the range.
        """
        if size < 0:
            msg = f"Expected size >= 0, but got start={start}, size={size}"
            _logger.error(msg)
            raise BufferReadError(msg)

        block = start // self._blocksize
        offset = start % self._blocksize

        # iterate over blocks in range and figure out long sub-ranges of blocks to fetch at once
        done = -offset
        to_fetch_start = None
        to_fetch_n = None
        assert self._cached_blocks is not None
        while done < size:
            if self._cached_blocks[block] is not None:
                # current block is loaded
                if to_fetch_start is not None:
                    # there was a block range to be loaded, do that now
                    self._fetch_blocks(to_fetch_start, to_fetch_n)

                    # no active block range anymore
                    to_fetch_start = None
                    to_fetch_n = None
            else:
                # current block is missing, do we already have a block range to append to?
                if to_fetch_start is None:
                    # no block range open, create a new one
                    to_fetch_start = block
                    to_fetch_n = 1
                else:
                    # current block range exists, append block
                    to_fetch_n += 1

            done += self._blocksize
            block += 1

        if to_fetch_start is not None:
            # this is the last active block range, fetch it
            self._fetch_blocks(to_fetch_start, to_fetch_n)

    def _read_data_from_blocks(self, start, size):
        """Read data from bytes.

        Parameters
        ----------
        start: int
            First byte of the range.
        size: int
            Number of bytes in the range.

        Returns
        -------
        data: bytes
            Requested data
        """
        block = start // self._blocksize
        offset = start % self._blocksize

        read_size = size + offset
        n_blocks = read_size // self._blocksize
        if read_size % self._blocksize != 0:
            n_blocks += 1

        assert self._cached_blocks is not None
        return b"".join(self._cached_blocks[block : (block + n_blocks)])[
            offset : (size + offset)
        ]

    def _check_closed(self):
        """Check that file object currently is not closed.

        Raises
        ------
        ValueError: in case file object is closed
        """
        if self.closed:
            raise ValueError("I/O operation on closed file.")

    def read(self, size=None):
        self._check_closed()
        self._setup_cache()

        if (size is None) or (size < 0) or (self._pos + size > self._size):
            # read entire, remaining file
            assert self._size is not None
            size = self._size - self._pos

        self._ensure_range_loaded(self._pos, size)
        result = self._read_data_from_blocks(self._pos, size)
        self._pos += size
        return result

    def tell(self):
        self._check_closed()
        return self._pos

    def seek(self, offset, whence=0):
        self._check_closed()
        self._setup_cache()

        if whence == 0:
            self._pos = max(0, min(offset, self._size))
        elif whence == 1:
            self._pos = max(0, min(self._pos + offset, self._size))
        elif whence == 2:
            self._pos = max(0, min(self._size + offset, self._size))
        else:
            raise ValueError("unsupported whence value")

        return self._pos

    @property
    def size(self):
        self._check_closed()
        self._setup_cache()

        return self._size

    def seekable(self):
        self._check_closed()
        return True

    def readable(self):
        self._check_closed()
        return True

    def close(self):
        if not self.closed:
            self._raw.close()
            super().close()
